# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt6 import QtCore
from PyQt6.QtWidgets import (QAbstractItemView, QCheckBox, QColorDialog, QComboBox, QDialog, QDialogButtonBox,
                             QFileDialog, QFormLayout, QFrame, QGridLayout, QGroupBox, QLabel, QLineEdit, QListWidget,
                             QListWidgetItem, QPushButton, QRadioButton, QScrollArea, QSpinBox, QSplitter,
                             QStackedWidget, QTabWidget, QVBoxLayout, QWidget)

from trackma import utils
from trackma.ui.qt.delegates import ShowsTableDelegate
from trackma.ui.qt.themedcolorpicker import ThemedColorPicker
from trackma.ui.qt.util import FilterBar, getColor, getIcon


class SettingsDialog(QDialog):
    worker = None
    config = None
    configfile = None

    saved = QtCore.pyqtSignal()

    def __init__(self, parent, worker, config, configfile):
        QDialog.__init__(self, parent)

        self.worker = worker
        self.config = config
        self.configfile = configfile
        self.setStyleSheet("QGroupBox { font-weight: bold; } ")
        self.setWindowTitle('Settings')
        layout = QGridLayout()

        # Categories
        self.category_list = QListWidget()
        category_media = QListWidgetItem(
            getIcon('media-playback-start'), 'Media', self.category_list)
        category_library = QListWidgetItem(
            getIcon('folder'), 'Library', self.category_list)
        category_sync = QListWidgetItem(
            getIcon('view-refresh'), 'Sync', self.category_list)
        category_ui = QListWidgetItem(
            getIcon('window-new'), 'User Interface', self.category_list)
        category_theme = QListWidgetItem(
            getIcon('applications-graphics'), 'Theme', self.category_list)
        self.category_list.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.category_list.setCurrentRow(0)
        self.category_list.setMaximumWidth(
            self.category_list.sizeHintForColumn(0) + 15)
        self.category_list.setFocus()
        self.category_list.currentItemChanged.connect(self.s_switch_page)

        # Media tab
        page_media = QWidget()
        page_media_layout = QVBoxLayout()
        page_media_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Group: Media settings
        g_media = QGroupBox('Media settings')
        g_media.setFlat(True)
        g_media_layout = QFormLayout()
        self.tracker_enabled = QCheckBox()
        self.tracker_enabled.toggled.connect(self.tracker_type_change)

        self.tracker_type = QComboBox()
        for (n, label) in utils.available_trackers:
            self.tracker_type.addItem(label, n)
        self.tracker_type.currentIndexChanged.connect(self.tracker_type_change)

        self.tracker_interval = QSpinBox()
        self.tracker_interval.setRange(5, 1000)
        self.tracker_interval.setMaximumWidth(60)
        self.tracker_process = QLineEdit()
        self.tracker_update_wait = QSpinBox()
        self.tracker_update_wait.setRange(0, 1000)
        self.tracker_update_wait.setMaximumWidth(60)
        self.tracker_update_close = QCheckBox()
        self.tracker_update_prompt = QCheckBox()
        self.tracker_not_found_prompt = QCheckBox()
        self.tracker_ignore_not_next = QCheckBox()

        g_media_layout.addRow('Enable tracker', self.tracker_enabled)
        g_media_layout.addRow('Tracker type', self.tracker_type)
        g_media_layout.addRow(
            'Tracker interval (seconds)', self.tracker_interval)
        g_media_layout.addRow('Process name (regex)', self.tracker_process)
        g_media_layout.addRow(
            'Wait before updating (seconds)', self.tracker_update_wait)
        g_media_layout.addRow(
            'Wait until the player is closed', self.tracker_update_close)
        g_media_layout.addRow('Ask before updating',
                              self.tracker_update_prompt)
        g_media_layout.addRow('Ask to add new shows',
                              self.tracker_not_found_prompt)
        g_media_layout.addRow('Ignore if not next episode',
                              self.tracker_ignore_not_next)

        g_media.setLayout(g_media_layout)

        # Group: Plex settings
        g_plex = QGroupBox('Plex Media Server')
        g_plex.setFlat(True)
        self.plex_host = QLineEdit()
        self.plex_port = QLineEdit()
        self.plex_user = QLineEdit()
        self.plex_passw = QLineEdit()
        self.plex_passw.setEchoMode(QLineEdit.EchoMode.Password)
        self.plex_obey_wait = QCheckBox()
        self.plex_ssl = QCheckBox()

        g_plex_layout = QGridLayout()
        g_plex_layout.addWidget(
            QLabel('Host and Port'),                   0, 0, 1, 1)
        g_plex_layout.addWidget(self.plex_host,
                                0, 1, 1, 1)
        g_plex_layout.addWidget(self.plex_port,
                                0, 2, 1, 2)
        g_plex_layout.addWidget(
            QLabel('Use SSL'), 1, 0, 1, 1)
        g_plex_layout.addWidget(self.plex_ssl,
                                1, 2, 1, 1)
        g_plex_layout.addWidget(
            QLabel('Use "wait before updating" time'), 2, 0, 1, 1)
        g_plex_layout.addWidget(self.plex_obey_wait,
                                2, 2, 1, 1)
        g_plex_layout.addWidget(
            QLabel('myPlex login (claimed server)'),   3, 0, 1, 1)
        g_plex_layout.addWidget(self.plex_user,
                                3, 1, 1, 1)
        g_plex_layout.addWidget(self.plex_passw,
                                3, 2, 1, 2)

        g_plex.setLayout(g_plex_layout)

        # Group: Jellyfin settings
        g_jellyfin = QGroupBox('Jellyfin')
        g_jellyfin.setFlat(True)
        self.jellyfin_host = QLineEdit()
        self.jellyfin_port = QLineEdit()
        self.jellyfin_user = QLineEdit()
        self.jellyfin_api_key = QLineEdit()

        g_jellyfin_layout = QGridLayout()
        g_jellyfin_layout.addWidget(
            QLabel('Host and Port'),                   0, 0, 1, 1)
        g_jellyfin_layout.addWidget(self.jellyfin_host,
                                0, 1, 1, 1)
        g_jellyfin_layout.addWidget(self.jellyfin_port,
                                0, 2, 1, 2)
        g_jellyfin_layout.addWidget(
            QLabel('API and User'),                   1, 0, 1, 1)
        g_jellyfin_layout.addWidget(self.jellyfin_api_key,
                                1, 1, 1, 1)
        g_jellyfin_layout.addWidget(self.jellyfin_user,
                                1, 2, 1, 2)

        g_jellyfin.setLayout(g_jellyfin_layout)

        # Group: Kodi settings
        g_kodi = QGroupBox('Kodi')
        g_kodi.setFlat(True)
        self.kodi_host = QLineEdit()
        self.kodi_port = QLineEdit()
        self.kodi_user = QLineEdit()
        self.kodi_passw = QLineEdit()
        self.kodi_passw.setEchoMode(QLineEdit.EchoMode.Password)
        self.kodi_obey_wait = QCheckBox()

        g_kodi_layout = QGridLayout()
        g_kodi_layout.addWidget(
            QLabel('Host and Port'),                   0, 0, 1, 1)
        g_kodi_layout.addWidget(self.kodi_host,
                                0, 1, 1, 1)
        g_kodi_layout.addWidget(self.kodi_port,
                                0, 2, 1, 2)
        g_kodi_layout.addWidget(
            QLabel('Use "wait before updating" time'), 1, 0, 1, 1)
        g_kodi_layout.addWidget(self.kodi_obey_wait,
                                1, 2, 1, 1)
        g_kodi_layout.addWidget(QLabel('Kodi login'),   2, 0, 1, 1)
        g_kodi_layout.addWidget(self.kodi_user,
                                2, 1, 1, 1)
        g_kodi_layout.addWidget(self.kodi_passw,
                                2, 2, 1, 2)

        g_kodi.setLayout(g_kodi_layout)

        # Media form
        page_media_layout.addWidget(g_media)
        page_media_layout.addWidget(g_plex)
        page_media_layout.addWidget(g_jellyfin)
        page_media_layout.addWidget(g_kodi)
        page_media.setLayout(page_media_layout)

        # Library tab
        page_library = QWidget()
        page_library_layout = QVBoxLayout()
        page_library_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Group: Library
        g_playnext = QGroupBox('Library')
        g_playnext.setFlat(True)
        self.title_parser = QComboBox()
        for (n, label) in utils.available_parsers:
            self.title_parser.addItem(label, n)
        self.player = QLineEdit()
        self.player_browse = QPushButton('Browse...')
        self.player_browse.clicked.connect(self.s_player_browse)
        lbl_searchdirs = QLabel('Media directories')
        lbl_searchdirs.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.searchdirs = QListWidget()
        self.searchdirs_add = QPushButton('Add...')
        self.searchdirs_add.clicked.connect(self.s_searchdirs_add)
        self.searchdirs_remove = QPushButton('Remove')
        self.searchdirs_remove.clicked.connect(self.s_searchdirs_remove)
        self.searchdirs_buttons = QVBoxLayout()
        self.searchdirs_buttons.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.searchdirs_buttons.addWidget(self.searchdirs_add)
        self.searchdirs_buttons.addWidget(self.searchdirs_remove)
        self.searchdirs_buttons.addWidget(QSplitter())
        self.library_autoscan = QCheckBox()
        self.scan_whole_list = QCheckBox()
        self.library_full_path = QCheckBox()

        g_playnext_layout = QGridLayout()
        g_playnext_layout.addWidget(
            QLabel('Parser'),                    0, 0, 1, 1)
        g_playnext_layout.addWidget(
            self.title_parser,                   0, 1, 1, 1)
        g_playnext_layout.addWidget(
            QLabel('Player'),                    1, 0, 1, 1)
        g_playnext_layout.addWidget(self.player,
                                                 1, 1, 1, 1)
        g_playnext_layout.addWidget(
            self.player_browse,                  1, 2, 1, 1)
        g_playnext_layout.addWidget(
            lbl_searchdirs,                      2, 0, 1, 1)
        g_playnext_layout.addWidget(
            self.searchdirs,                     2, 1, 1, 1)
        g_playnext_layout.addLayout(
            self.searchdirs_buttons,             2, 2, 1, 1)
        g_playnext_layout.addWidget(
            QLabel('Rescan Library at startup'), 3, 0, 1, 2)
        g_playnext_layout.addWidget(
            self.library_autoscan,               3, 2, 1, 1)
        g_playnext_layout.addWidget(
            QLabel('Scan through whole list'),   4, 0, 1, 2)
        g_playnext_layout.addWidget(
            self.scan_whole_list,                4, 2, 1, 1)
        g_playnext_layout.addWidget(
            QLabel('Take subdirectory name into account'), 5, 0, 1, 2)
        g_playnext_layout.addWidget(
            self.library_full_path,              5, 2, 1, 1)

        g_playnext.setLayout(g_playnext_layout)

        page_library_layout.addWidget(g_playnext)
        page_library.setLayout(page_library_layout)

        # Sync tab
        page_sync = QWidget()
        page_sync_layout = QVBoxLayout()
        page_sync_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Group: Autoretrieve
        g_autoretrieve = QGroupBox('Autoretrieve')
        g_autoretrieve.setFlat(True)
        self.autoretrieve_off = QRadioButton('Disabled')
        self.autoretrieve_always = QRadioButton('Always at start')
        self.autoretrieve_days = QRadioButton('After n days')
        self.autoretrieve_days.toggled.connect(self.s_autoretrieve_days)
        self.autoretrieve_days_n = QSpinBox()
        self.autoretrieve_days_n.setRange(1, 100)
        g_autoretrieve_layout = QGridLayout()
        g_autoretrieve_layout.setColumnStretch(0, 1)
        g_autoretrieve_layout.addWidget(self.autoretrieve_off,    0, 0, 1, 1)
        g_autoretrieve_layout.addWidget(self.autoretrieve_always, 1, 0, 1, 1)
        g_autoretrieve_layout.addWidget(self.autoretrieve_days,   2, 0, 1, 1)
        g_autoretrieve_layout.addWidget(self.autoretrieve_days_n, 2, 1, 1, 1)
        g_autoretrieve.setLayout(g_autoretrieve_layout)

        # Group: Autosend
        g_autosend = QGroupBox('Autosend')
        g_autosend.setFlat(True)
        self.autosend_off = QRadioButton('Disabled')
        self.autosend_always = QRadioButton('Immediately after every change')
        self.autosend_minutes = QRadioButton('After n minutes')
        self.autosend_minutes.toggled.connect(self.s_autosend_minutes)
        self.autosend_minutes_n = QSpinBox()
        self.autosend_minutes_n.setRange(1, 1000)
        self.autosend_size = QRadioButton('After the queue reaches n items')
        self.autosend_size.toggled.connect(self.s_autosend_size)
        self.autosend_size_n = QSpinBox()
        self.autosend_size_n.setRange(2, 20)
        self.autosend_at_exit = QCheckBox('At exit')
        g_autosend_layout = QGridLayout()
        g_autosend_layout.setColumnStretch(0, 1)
        g_autosend_layout.addWidget(self.autosend_off,      0, 0, 1, 1)
        g_autosend_layout.addWidget(self.autosend_always,   1, 0, 1, 1)
        g_autosend_layout.addWidget(self.autosend_minutes,    2, 0, 1, 1)
        g_autosend_layout.addWidget(self.autosend_minutes_n,  2, 1, 1, 1)
        g_autosend_layout.addWidget(self.autosend_size,     3, 0, 1, 1)
        g_autosend_layout.addWidget(self.autosend_size_n,   3, 1, 1, 1)
        g_autosend_layout.addWidget(self.autosend_at_exit,  4, 0, 1, 1)
        g_autosend.setLayout(g_autosend_layout)

        # Group: Extra
        g_extra = QGroupBox('Additional options')
        g_extra.setFlat(True)
        self.auto_status_change = QCheckBox('Change status automatically')
        self.auto_status_change.toggled.connect(self.s_auto_status_change)
        self.auto_status_change_if_scored = QCheckBox(
            'Change status automatically only if scored')
        self.auto_date_change = QCheckBox(
            'Change start and finish dates automatically')
        g_extra_layout = QVBoxLayout()
        g_extra_layout.addWidget(self.auto_status_change)
        g_extra_layout.addWidget(self.auto_status_change_if_scored)
        g_extra_layout.addWidget(self.auto_date_change)
        g_extra.setLayout(g_extra_layout)

        # Sync layout
        page_sync_layout.addWidget(g_autoretrieve)
        page_sync_layout.addWidget(g_autosend)
        page_sync_layout.addWidget(g_extra)
        page_sync.setLayout(page_sync_layout)

        # UI tab
        page_ui = QWidget()
        page_ui_layout = QFormLayout()
        page_ui_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Group: Icon
        g_icon = QGroupBox('Notification Icon')
        g_icon.setFlat(True)
        self.tray_icon = QCheckBox('Show tray icon')
        self.tray_icon.toggled.connect(self.s_tray_icon)
        self.close_to_tray = QCheckBox('Close to tray')
        self.start_in_tray = QCheckBox('Start minimized to tray')
        self.tray_api_icon = QCheckBox('Use API icon as tray icon')
        self.notifications = QCheckBox(
            'Show notification when tracker detects new media')
        g_icon_layout = QVBoxLayout()
        g_icon_layout.addWidget(self.tray_icon)
        g_icon_layout.addWidget(self.close_to_tray)
        g_icon_layout.addWidget(self.start_in_tray)
        g_icon_layout.addWidget(self.tray_api_icon)
        g_icon_layout.addWidget(self.notifications)
        g_icon.setLayout(g_icon_layout)

        # Group: Window
        g_window = QGroupBox('Window')
        g_window.setFlat(True)
        self.remember_geometry = QCheckBox('Remember window size and position')
        self.remember_columns = QCheckBox('Remember column layouts and widths')
        self.columns_per_api = QCheckBox(
            'Use different visible columns per API')
        g_window_layout = QVBoxLayout()
        g_window_layout.addWidget(self.remember_geometry)
        g_window_layout.addWidget(self.remember_columns)
        g_window_layout.addWidget(self.columns_per_api)
        g_window.setLayout(g_window_layout)

        # Group: Lists
        g_lists = QGroupBox('Lists')
        g_lists.setFlat(True)
        self.filter_bar_position = QComboBox()
        filter_bar_positions = [(FilterBar.PositionHidden,     'Hidden'),
                                (FilterBar.PositionAboveLists, 'Above lists'),
                                (FilterBar.PositionBelowLists, 'Below lists')]
        for (n, label) in filter_bar_positions:
            self.filter_bar_position.addItem(label, n)
        self.inline_edit = QCheckBox('Enable in-line editing')
        g_lists_layout = QFormLayout()
        g_lists_layout.addRow('Filter bar position:', self.filter_bar_position)
        g_lists_layout.addRow(self.inline_edit)
        g_lists.setLayout(g_lists_layout)

        # UI layout
        page_ui_layout.addWidget(g_icon)
        page_ui_layout.addWidget(g_window)
        page_ui_layout.addWidget(g_lists)
        page_ui.setLayout(page_ui_layout)

        # Theming tab
        page_theme = QWidget()
        page_theme_layout = QFormLayout()
        page_theme_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Group: Episode Bar
        g_ep_bar = QGroupBox('Episode Bar')
        g_ep_bar.setFlat(True)
        self.ep_bar_style = QComboBox()
        ep_bar_styles = [(ShowsTableDelegate.BarStyleBasic,  'Basic'),
                         (ShowsTableDelegate.BarStyle04,     'Trackma'),
                         (ShowsTableDelegate.BarStyleHybrid, 'Hybrid')]
        for (n, label) in ep_bar_styles:
            self.ep_bar_style.addItem(label, n)
        self.ep_bar_style.currentIndexChanged.connect(self.s_ep_bar_style)
        self.ep_bar_text = QCheckBox('Show text label')
        g_ep_bar_layout = QFormLayout()
        g_ep_bar_layout.addRow('Style:', self.ep_bar_style)
        g_ep_bar_layout.addRow(self.ep_bar_text)
        g_ep_bar.setLayout(g_ep_bar_layout)

        # Group: Colour scheme
        g_scheme = QGroupBox('Color Scheme')
        g_scheme.setFlat(True)
        col_tabs = [('rows',     '&Row highlights'),
                    ('progress', '&Progress widget')]
        self.colors = {}
        self.colors['rows'] = [('is_playing',  'Playing'),
                               ('is_queued',   'Queued'),
                               ('new_episode', 'New Episode'),
                               ('is_airing',   'Airing'),
                               ('not_aired',   'Unaired')]
        self.colors['progress'] = [('progress_bg',       'Background'),
                                   ('progress_fg',       'Watched bar'),
                                   ('progress_sub_bg',   'Aired episodes'),
                                   ('progress_sub_fg',   'Stored episodes'),
                                   ('progress_complete', 'Complete')]
        self.color_buttons = []
        self.syscolor_buttons = []
        g_scheme_layout = QGridLayout()
        tw_scheme = QTabWidget()
        for (key1, tab_title) in col_tabs:
            page = QFrame()
            page_layout = QGridLayout()
            col = 0
            # Generate widgets from the keys and values
            for (key2, label) in self.colors[key1]:
                self.color_buttons.append(QPushButton())
                # self.color_buttons[-1].setStyleSheet('background-color: ' + getColor(self.config['colors'][key]).name())
                self.color_buttons[-1].setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.color_buttons[-1].clicked.connect(
                    self.s_color_picker(key2, False))
                self.syscolor_buttons.append(QPushButton('System Colors'))
                self.syscolor_buttons[-1].clicked.connect(
                    self.s_color_picker(key2, True))
                page_layout.addWidget(QLabel(label),             col, 0, 1, 1)
                page_layout.addWidget(self.color_buttons[-1],    col, 1, 1, 1)
                page_layout.addWidget(self.syscolor_buttons[-1], col, 2, 1, 1)
                col += 1
            page.setLayout(page_layout)
            tw_scheme.addTab(page, tab_title)
        g_scheme_layout.addWidget(tw_scheme)
        g_scheme.setLayout(g_scheme_layout)

        # UI layout
        page_theme_layout.addWidget(g_ep_bar)
        page_theme_layout.addWidget(g_scheme)
        page_theme.setLayout(page_theme_layout)

        # Content
        self.contents = QStackedWidget()
        for page in (page_media, page_library, page_sync, page_ui, page_theme):
            scrollable_page = QScrollArea()
            scrollable_page.setWidgetResizable(True)
            scrollable_page.setWidget(page)
            self.contents.addWidget(scrollable_page)

        # Bottom buttons
        bottombox = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Apply
            | QDialogButtonBox.StandardButton.Cancel
        )
        bottombox.accepted.connect(self.s_save)
        bottombox.button(QDialogButtonBox.StandardButton.Apply).clicked.connect(self._save)
        bottombox.rejected.connect(self.reject)

        # Main layout finish
        layout.addWidget(self.category_list,  0, 0, 1, 1)
        layout.addWidget(self.contents,       0, 1, 1, 1)
        layout.addWidget(bottombox,           1, 0, 1, 2)
        layout.setColumnStretch(1, 1)

        self._load()
        self.update_colors()

        self.setLayout(layout)

    def _add_dir(self, path):
        self.searchdirs.addItem(path)

    def _load(self):
        engine = self.worker.engine
        tracker_type = self.tracker_type.findData(
            engine.get_config('tracker_type'))
        title_parser = self.title_parser.findData(
            engine.get_config('title_parser'))
        autoretrieve = engine.get_config('autoretrieve')
        autosend = engine.get_config('autosend')

        self.tracker_enabled.setChecked(engine.get_config('tracker_enabled'))
        self.tracker_type.setCurrentIndex(max(0, tracker_type))
        self.tracker_interval.setValue(engine.get_config('tracker_interval'))
        self.tracker_process.setText(engine.get_config('tracker_process'))
        self.tracker_update_wait.setValue(
            engine.get_config('tracker_update_wait_s'))
        self.tracker_update_close.setChecked(
            engine.get_config('tracker_update_close'))
        self.tracker_update_prompt.setChecked(
            engine.get_config('tracker_update_prompt'))
        self.tracker_not_found_prompt.setChecked(
            engine.get_config('tracker_not_found_prompt'))
        self.tracker_ignore_not_next.setChecked(
            engine.get_config('tracker_ignore_not_next'))

        self.title_parser.setCurrentIndex(max(0, title_parser))
        self.player.setText(engine.get_config('player'))
        self.library_autoscan.setChecked(engine.get_config('library_autoscan'))
        self.scan_whole_list.setChecked(engine.get_config('scan_whole_list'))
        self.library_full_path.setChecked(
            engine.get_config('library_full_path'))

        self.plex_host.setText(engine.get_config('plex_host'))
        self.plex_port.setText(engine.get_config('plex_port'))
        self.plex_ssl.setChecked(
            engine.get_config('plex_ssl'))
        self.plex_obey_wait.setChecked(
            engine.get_config('plex_obey_update_wait_s'))
        self.plex_user.setText(engine.get_config('plex_user'))
        self.plex_passw.setText(engine.get_config('plex_passwd'))

        self.jellyfin_host.setText(engine.get_config('jellyfin_host'))
        self.jellyfin_port.setText(engine.get_config('jellyfin_port'))
        self.jellyfin_user.setText(engine.get_config('jellyfin_user'))
        self.jellyfin_api_key.setText(engine.get_config('jellyfin_api_key'))

        self.kodi_host.setText(engine.get_config('kodi_host'))
        self.kodi_port.setText(engine.get_config('kodi_port'))
        self.kodi_obey_wait.setChecked(
            engine.get_config('kodi_obey_update_wait_s'))
        self.kodi_user.setText(engine.get_config('kodi_user'))
        self.kodi_passw.setText(engine.get_config('kodi_passwd'))

        for path in engine.get_config('searchdir'):
            self._add_dir(path)

        if autoretrieve == 'always':
            self.autoretrieve_always.setChecked(True)
        elif autoretrieve == 'days':
            self.autoretrieve_days.setChecked(True)
        else:
            self.autoretrieve_off.setChecked(True)

        self.autoretrieve_days_n.setValue(
            engine.get_config('autoretrieve_days'))

        if autosend == 'always':
            self.autosend_always.setChecked(True)
        elif autosend in ('minutes', 'hours'):
            self.autosend_minutes.setChecked(True)
        elif autosend == 'size':
            self.autosend_size.setChecked(True)
        else:
            self.autosend_off.setChecked(True)

        self.autosend_minutes_n.setValue(engine.get_config('autosend_minutes'))
        self.autosend_size_n.setValue(engine.get_config('autosend_size'))

        self.autosend_at_exit.setChecked(engine.get_config('autosend_at_exit'))
        self.auto_status_change.setChecked(
            engine.get_config('auto_status_change'))
        self.auto_status_change_if_scored.setChecked(
            engine.get_config('auto_status_change_if_scored'))
        self.auto_date_change.setChecked(engine.get_config('auto_date_change'))

        self.tray_icon.setChecked(self.config['show_tray'])
        self.close_to_tray.setChecked(self.config['close_to_tray'])
        self.start_in_tray.setChecked(self.config['start_in_tray'])
        self.tray_api_icon.setChecked(self.config['tray_api_icon'])
        self.notifications.setChecked(self.config['notifications'])
        self.remember_geometry.setChecked(self.config['remember_geometry'])
        self.remember_columns.setChecked(self.config['remember_columns'])
        self.columns_per_api.setChecked(self.config['columns_per_api'])
        self.filter_bar_position.setCurrentIndex(
            self.filter_bar_position.findData(self.config['filter_bar_position']))
        self.inline_edit.setChecked(self.config['inline_edit'])

        self.ep_bar_style.setCurrentIndex(
            self.ep_bar_style.findData(self.config['episodebar_style']))
        self.ep_bar_text.setChecked(self.config['episodebar_text'])

        self.autoretrieve_days_n.setEnabled(self.autoretrieve_days.isChecked())
        self.autosend_minutes_n.setEnabled(self.autosend_minutes.isChecked())
        self.autosend_size_n.setEnabled(self.autosend_size.isChecked())
        self.close_to_tray.setEnabled(self.tray_icon.isChecked())
        self.start_in_tray.setEnabled(self.tray_icon.isChecked())
        self.notifications.setEnabled(self.tray_icon.isChecked())

        self.color_values = self.config['colors'].copy()

        self.tracker_type_change(None)

    def _save(self):
        engine = self.worker.engine

        engine.set_config('tracker_enabled',
                          self.tracker_enabled.isChecked())
        engine.set_config('tracker_type',          self.tracker_type.itemData(
            self.tracker_type.currentIndex()))
        engine.set_config('tracker_interval',
                          self.tracker_interval.value())
        engine.set_config('tracker_process',       str(
            self.tracker_process.text()))
        engine.set_config('tracker_update_wait_s',
                          self.tracker_update_wait.value())
        engine.set_config('tracker_update_close',
                          self.tracker_update_close.isChecked())
        engine.set_config('tracker_update_prompt',
                          self.tracker_update_prompt.isChecked())
        engine.set_config('tracker_not_found_prompt',
                          self.tracker_not_found_prompt.isChecked())
        engine.set_config('tracker_ignore_not_next',
                          self.tracker_ignore_not_next.isChecked())

        engine.set_config('title_parser',          self.title_parser.itemData(
            self.title_parser.currentIndex()))
        engine.set_config('player',            self.player.text())
        engine.set_config('library_autoscan',
                          self.library_autoscan.isChecked())
        engine.set_config('scan_whole_list', self.scan_whole_list.isChecked())
        engine.set_config('library_full_path',
                          self.library_full_path.isChecked())

        engine.set_config('plex_host',         self.plex_host.text())
        engine.set_config('plex_port',         self.plex_port.text())
        engine.set_config('plex_obey_update_wait_s',
                          self.plex_obey_wait.isChecked())
        engine.set_config('plex_ssl',          self.plex_ssl.isChecked())
        engine.set_config('plex_user',         self.plex_user.text())
        engine.set_config('plex_passwd',       self.plex_passw.text())

        engine.set_config('jellyfin_host',         self.jellyfin_host.text())
        engine.set_config('jellyfin_port',         self.jellyfin_port.text())
        engine.set_config('jellyfin_user',         self.jellyfin_user.text())
        engine.set_config('jellyfin_api_key',       self.jellyfin_api_key.text())

        engine.set_config('kodi_host',         self.kodi_host.text())
        engine.set_config('kodi_port',         self.kodi_port.text())
        engine.set_config('kodi_obey_update_wait_s',
                          self.kodi_obey_wait.isChecked())
        engine.set_config('kodi_user',         self.kodi_user.text())
        engine.set_config('kodi_passwd',       self.kodi_passw.text())

        engine.set_config('searchdir',         [self.searchdirs.item(
            i).text() for i in range(self.searchdirs.count())])

        if self.autoretrieve_always.isChecked():
            engine.set_config('autoretrieve', 'always')
        elif self.autoretrieve_days.isChecked():
            engine.set_config('autoretrieve', 'days')
        else:
            engine.set_config('autoretrieve', 'off')

        engine.set_config('autoretrieve_days',
                          self.autoretrieve_days_n.value())

        if self.autosend_always.isChecked():
            engine.set_config('autosend', 'always')
        elif self.autosend_minutes.isChecked():
            engine.set_config('autosend', 'minutes')
        elif self.autosend_size.isChecked():
            engine.set_config('autosend', 'size')
        else:
            engine.set_config('autosend', 'off')

        engine.set_config('autosend_minutes', self.autosend_minutes_n.value())
        engine.set_config('autosend_size',  self.autosend_size_n.value())

        engine.set_config('autosend_at_exit',
                          self.autosend_at_exit.isChecked())
        engine.set_config('auto_status_change',
                          self.auto_status_change.isChecked())
        engine.set_config('auto_status_change_if_scored',
                          self.auto_status_change_if_scored.isChecked())
        engine.set_config('auto_date_change',
                          self.auto_date_change.isChecked())

        engine.save_config()

        self.config['show_tray'] = self.tray_icon.isChecked()
        self.config['close_to_tray'] = self.close_to_tray.isChecked()
        self.config['start_in_tray'] = self.start_in_tray.isChecked()
        self.config['tray_api_icon'] = self.tray_api_icon.isChecked()
        self.config['notifications'] = self.notifications.isChecked()
        self.config['remember_geometry'] = self.remember_geometry.isChecked()
        self.config['remember_columns'] = self.remember_columns.isChecked()
        self.config['columns_per_api'] = self.columns_per_api.isChecked()
        self.config['filter_bar_position'] = self.filter_bar_position.itemData(
            self.filter_bar_position.currentIndex())
        self.config['inline_edit'] = self.inline_edit.isChecked()

        self.config['episodebar_style'] = self.ep_bar_style.itemData(
            self.ep_bar_style.currentIndex())
        self.config['episodebar_text'] = self.ep_bar_text.isChecked()

        self.config['colors'] = self.color_values

        utils.save_config(self.config, self.configfile)

        self.saved.emit()

    def s_save(self):
        self._save()
        self.accept()

    def switch_kodi_state(self, state):
        self.kodi_host.setEnabled(state)
        self.kodi_port.setEnabled(state)
        self.kodi_user.setEnabled(state)
        self.kodi_passw.setEnabled(state)
        self.kodi_obey_wait.setEnabled(state)

    def switch_plex_state(self, state):
        self.plex_host.setEnabled(state)
        self.plex_port.setEnabled(state)
        self.plex_user.setEnabled(state)
        self.plex_passw.setEnabled(state)
        self.plex_obey_wait.setEnabled(state)
        self.plex_ssl.setEnabled(state)

    def switch_jellyfin_state(self, state):
        self.jellyfin_host.setEnabled(state)
        self.jellyfin_port.setEnabled(state)
        self.jellyfin_user.setEnabled(state)
        self.jellyfin_api_key.setEnabled(state)

    def tracker_type_change(self, checked):
        if self.tracker_enabled.isChecked():
            self.tracker_interval.setEnabled(True)
            self.tracker_update_wait.setEnabled(True)
            self.tracker_type.setEnabled(True)
            if self.tracker_type.itemData(self.tracker_type.currentIndex()) == 'plex':
                self.switch_jellyfin_state(False)
                self.switch_plex_state(True)
                self.switch_kodi_state(False)
                self.tracker_process.setEnabled(False)
            elif self.tracker_type.itemData(self.tracker_type.currentIndex()) == 'jellyfin':
                self.switch_jellyfin_state(True)
                self.switch_kodi_state(False)
                self.switch_plex_state(False)
                self.tracker_process.setEnabled(False)
            elif self.tracker_type.itemData(self.tracker_type.currentIndex()) == 'kodi':
                self.switch_jellyfin_state(False)
                self.switch_kodi_state(True)
                self.switch_plex_state(False)
                self.tracker_process.setEnabled(False)
            else:
                self.tracker_process.setEnabled(True)
                self.switch_jellyfin_state(False)
                self.switch_plex_state(False)
                self.switch_kodi_state(False)
        else:
            self.tracker_type.setEnabled(False)
            self.switch_jellyfin_state(False)
            self.switch_plex_state(False)
            self.switch_kodi_state(False)

            self.tracker_process.setEnabled(False)
            self.tracker_interval.setEnabled(False)
            self.tracker_update_wait.setEnabled(False)

    def s_autoretrieve_days(self, checked):
        self.autoretrieve_days_n.setEnabled(checked)

    def s_autosend_minutes(self, checked):
        self.autosend_minutes_n.setEnabled(checked)

    def s_autosend_size(self, checked):
        self.autosend_size_n.setEnabled(checked)

    def s_tray_icon(self, checked):
        self.close_to_tray.setEnabled(checked)
        self.start_in_tray.setEnabled(checked)
        self.tray_api_icon.setEnabled(checked)
        self.notifications.setEnabled(checked)

    def s_ep_bar_style(self, index):
        if self.ep_bar_style.itemData(index) == ShowsTableDelegate.BarStyle04:
            self.ep_bar_text.setEnabled(False)
        else:
            self.ep_bar_text.setEnabled(True)

    def s_auto_status_change(self, checked):
        self.auto_status_change_if_scored.setEnabled(checked)

    def s_player_browse(self):
        self.player.setText(QFileDialog.getOpenFileName(
            caption='Choose player executable')[0])

    def s_searchdirs_add(self):
        self._add_dir(QFileDialog.getExistingDirectory(
            caption='Choose media directory'))

    def s_searchdirs_remove(self):
        row = self.searchdirs.currentRow()
        if row != -1:
            self.searchdirs.takeItem(row)

    def s_switch_page(self, new, old):
        if not new:
            new = old

        self.contents.setCurrentIndex(self.category_list.row(new))

    def s_color_picker(self, key, system):
        return lambda: self.color_picker(key, system)

    def color_picker(self, key, system):
        if system is True:
            current = self.color_values[key]
            result = ThemedColorPicker.do()
            if result is not None and result is not current:
                self.color_values[key] = result
                self.update_colors()
        else:
            current = getColor(self.color_values[key])
            result = QColorDialog.getColor(current)
            if result.isValid() and result is not current:
                self.color_values[key] = str(result.name())
                self.update_colors()

    def update_colors(self):
        for ((key, label), color) in zip(self.colors['rows']+self.colors['progress'], self.color_buttons):
            color.setStyleSheet('background-color: ' +
                                getColor(self.color_values[key]).name())



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

SUPPORTED_APIS = ['mal', 'kitsu', 'anilist']
SUPPORTED_MEDIATYPES = ['anime']


def supports(api, mediatype):
    return api in SUPPORTED_APIS and mediatype in SUPPORTED_MEDIATYPES


def parse_anime_relations(filename, api, last=None):
    """
    Support for Taiga-style anime relations file.
    Thanks to erengy and all the contributors.
    Database under the public domain.

    https://github.com/erengy/anime-relations
    """
    (src_grp, dst_grp) = (SUPPORTED_APIS.index(api) + 1, SUPPORTED_APIS.index(api) + 6)

    with open(filename) as f:
        import re

        relations = {'meta': {}}

        id_pattern = r"(\d+|[\?~])\|(\d+|[\?~])\|(\d+|[\?~])"
        ep_pattern = r"(\d+)-?(\d+|\?)?"
        full = r'- {0}:{1} -> {0}:{1}(!)?'.format(id_pattern, ep_pattern)
        _re = re.compile(full)

        mode = 0

        for line in f:
            line = line.strip()

            if not line:
                continue
            if line[0] == '#':
                continue

            if mode == 0 and line == "::meta":
                mode = 1
            elif mode == 1:
                if line[:16] == "- last_modified:":
                    last_modified = line[17:]

                    # TODO : Stop if the file hasn't changed
                    if last and last == last_modified:
                        return None

                    relations['meta']['last_modified'] = last_modified
                elif line == "::rules":
                    mode = 2
            elif mode == 2 and line[0] == '-':
                m = _re.match(line)
                if m:
                    # Source
                    src_id = m.group(src_grp)

                    # Handle unknown IDs
                    if src_id == '?':
                        continue
                    else:
                        src_id = int(src_id)

                    # Handle infinite ranges
                    if m.group(5) == '?':
                        src_eps = (int(m.group(4)), -1)
                    else:
                        src_eps = (int(m.group(4)), int(
                            m.group(5) or m.group(4)))

                    # Destination
                    dst_id = m.group(dst_grp)

                    # Handle ID repeaters
                    if dst_id == '~':
                        dst_id = src_id
                    else:
                        dst_id = int(dst_id)

                    # Handle infinite ranges
                    if m.group(10) == '?':
                        dst_eps = (int(m.group(9)), -1)
                    else:
                        dst_eps = (int(m.group(9)), int(
                            m.group(10) or m.group(9)))

                    if src_id not in relations:
                        relations[src_id] = []
                    relations[src_id].append((src_eps, dst_id, dst_eps))
                    # Handle self-redirecting rules
                    if m.group(11) == '!':
                        if dst_id not in relations:
                            relations[dst_id] = []
                        relations[dst_id].append((src_eps, dst_id, dst_eps))
                else:
                    print("Not recognized. " + line)

        return relations



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import html
import os
import threading

from gi.repository import GObject, Gtk

from trackma import utils
from trackma.ui.gtk import gtk_dir
from trackma.ui.gtk.imagebox import ImageBox


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/showinfobox.ui'))
class ShowInfoBox(Gtk.Box):
    __gtype_name__ = 'ShowInfoBox'

    label_title = Gtk.Template.Child()
    data_container = Gtk.Template.Child()
    image_container = Gtk.Template.Child()

    def __init__(self, engine, orientation=Gtk.Orientation.HORIZONTAL):
        Gtk.Box.__init__(self)
        self.init_template()

        self._engine = engine
        self._show = None
        self.image_thread = None
        self.details = None
        self.details_e = None

        self.image_box = ImageBox(225, 300)
        self.image_box.show()
        self.image_container.pack_start(self.image_box, False, False, 0)

        self.data_label = Gtk.Label('')
        self.data_label.set_line_wrap(True)
        self.data_label.set_property('selectable', True)

        if isinstance(orientation, Gtk.Orientation):
            self.data_container.set_orientation(orientation)
        self.data_container.pack_start(self.data_label, True, True, 0)

    def set_size(self, w, h):
        self.scrolled_sidebox.set_size_request(w, h)

    def load(self, show):
        self._show = show

        # Load image
        if show.get('image'):
            utils.make_dir(utils.to_cache_path())
            imagefile = utils.to_cache_path("%s_%s_f_%s.jpg" % (self._engine.api_info['shortname'],
                                                                self._engine.api_info['mediatype'],
                                                                show['id']))

            if os.path.isfile(imagefile):
                self.image_box.set_image(imagefile)
            else:
                self.image_box.set_image_remote(show['image'], imagefile)
        else:
            self.image_box.set_text('No Image')

        # Start info loading thread
        threading.Thread(target=self._show_load_start_task).start()

    def _show_load_start_task(self):
        # Thread to ask the engine for show details
        try:
            self.details = self._engine.get_show_details(self._show)
        except utils.TrackmaError as e:
            self.details = None
            self.details_e = e

        GObject.idle_add(self._show_load_finish_idle)

    def _show_load_finish_idle(self):
        if self.details:
            # Put the returned details into the lines VBox
            self.label_title.set_text(html.escape(self.details['title']))

            detail = list()
            for line in self.details['extra']:
                if line[0] and line[1]:
                    title, content, *_ = line

                    if isinstance(content, list):
                        content = ", ".join(filter(None, content))

                    detail.append("<b>%s</b>\n%s" % (html.escape(str(title)),
                                                     html.escape(str(content))))

            self.data_label.set_text("\n\n".join(detail))
            self.data_label.set_use_markup(True)
        else:
            self.label_title.set_text('Error while getting details.')

            if self.details_e:
                self.data_label.set_text(str(self.details_e))

        self.label_title.show()
        self.data_label.show()



import datetime

from PyQt6 import QtCore, QtGui

from trackma import utils
from trackma.ui.qt.thumbs import ThumbManager
from trackma.ui.qt.util import getColor, getIcon


class ShowListModel(QtCore.QAbstractTableModel):
    """
    Main model used in the main window to show
    a list of shows in the user's list.
    """
    COL_ID = 0
    COL_TITLE = 1
    COL_MY_PROGRESS = 2
    COL_MY_SCORE = 3
    COL_PERCENT = 4
    COL_NEXT_EP = 5
    COL_START_DATE = 6
    COL_END_DATE = 7
    COL_MY_START = 8
    COL_MY_FINISH = 9
    COL_MY_TAGS = 10
    COL_MY_STATUS = 11
    COL_LAST_UPDATED = 12

    columns = ['ID', 'Title', 'Progress', 'Score',
               'Percent', 'Next Episode', 'Start date', 'End date',
               'My start', 'My finish', 'Tags', 'Status', 'Last updated']

    editable_columns = [COL_MY_PROGRESS, COL_MY_SCORE]

    common_flags = \
        QtCore.Qt.ItemFlag.ItemIsSelectable | \
        QtCore.Qt.ItemFlag.ItemIsEnabled | \
        QtCore.Qt.ItemFlag.ItemNeverHasChildren

    date_format = "%Y-%m-%d"

    progressChanged = QtCore.pyqtSignal(QtCore.QVariant, float)
    scoreChanged = QtCore.pyqtSignal(QtCore.QVariant, float)

    def __init__(self, parent=None, palette=None):
        self.showlist = None
        self.palette = palette
        self.playing = set()
        self.mediainfo = {}

        super().__init__(parent)

    def setDateFormat(self, date_format):
        self.date_format = date_format

    def setMediaInfo(self, mediainfo):
        self.mediainfo = mediainfo

    def _date(self, obj):
        if obj:
            return obj.strftime(self.date_format)
        else:
            return '-'

    def _calculate_color(self, row, show):
        color = None

        if show['id'] in self.playing:
            color = 'is_playing'
        elif show.get('queued'):
            color = 'is_queued'
        elif self.library.get(show['id']) and max(self.library.get(show['id'])) > show['my_progress']:
            color = 'new_episode'
        elif show['status'] == utils.Status.AIRING:
            color = 'is_airing'
        elif show['status'] == utils.Status.NOTYET:
            color = 'not_aired'
        else:
            color = None

        if color:
            self.colors[row] = QtGui.QBrush(getColor(self.palette[color]))
        elif row in self.colors:
            del self.colors[row]

    def _calculate_next_ep(self, row, show):
        if self.mediainfo.get('date_next_ep'):
            if 'next_ep_time' in show:
                delta = show['next_ep_time'] - datetime.datetime.utcnow()
                self.next_ep[row] = "%i days, %02d hrs." % (
                    delta.days, delta.seconds/3600)
            elif row in self.next_ep:
                del self.next_ep[row]

    def _calculate_eps(self, row, show):
        aired_eps = utils.estimate_aired_episodes(show)
        library_eps = self.library.get(show['id'])

        if library_eps:
            library_eps = library_eps.keys()

        if aired_eps or library_eps:
            self.eps[row] = (aired_eps, library_eps)
        elif row in self.eps:
            del self.eps[row]

    def setShowList(self, showlist, altnames, library):
        self.beginResetModel()

        self.showlist = list(showlist)
        self.altnames = altnames
        self.library = library

        self.id_map = {}
        self.colors = {}
        self.next_ep = {}
        self.eps = {}

        for row, show in enumerate(self.showlist):
            self.id_map[show['id']] = row
            self._calculate_color(row, show)

            if self.mediainfo.get('can_play'):
                self._calculate_next_ep(row, show)
                self._calculate_eps(row, show)

        self.endResetModel()

    def update(self, showid, is_playing=None):
        if not self.showlist:
            return

        # Recalculate color and emit the changed signal
        row = self.id_map[showid]
        show = self.showlist[row]

        if is_playing is not None:
            if is_playing:
                self.playing.add(showid)
            else:
                self.playing.discard(showid)

        self._calculate_color(row, show)
        self.dataChanged.emit(self.index(
            row, 0), self.index(row, len(self.columns)-1))

    def rowCount(self, parent=QtCore.QModelIndex()):
        if self.showlist:
            return len(self.showlist)
        else:
            return 0

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self.columns)

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole and orientation == QtCore.Qt.Orientation.Horizontal:
            return self.columns[section]
        elif role == QtCore.Qt.ItemDataRole.ToolTipRole and orientation == QtCore.Qt.Orientation.Horizontal:
            if section == ShowListModel.COL_LAST_UPDATED:
                return 'Date and time of the last synced update'

    def setData(self, index, value, role):
        row, column = index.row(), index.column()
        show = self.showlist[row]

        if column == ShowListModel.COL_MY_PROGRESS:
            self.progressChanged.emit(show['id'], value)
        elif column == ShowListModel.COL_MY_SCORE:
            self.scoreChanged.emit(show['id'], value)

        return True

    def data(self, index, role):
        row, column = index.row(), index.column()
        show = self.showlist[row]

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if column == ShowListModel.COL_ID:
                return show['id']
            elif column == ShowListModel.COL_TITLE:
                title_str = show['title']
                if show['id'] in self.altnames:
                    title_str += " [%s]" % self.altnames[show['id']]
                return title_str
            elif column == ShowListModel.COL_MY_PROGRESS:
                return "{} / {}".format(show['my_progress'], show['total'] or '?')
            elif column == ShowListModel.COL_MY_SCORE:
                return show['my_score']
            elif column == ShowListModel.COL_PERCENT:
                # return "{:.0%}".format(show['my_progress'] / 100)
                if show['total']:
                    total = show['total']
                else:
                    total = (int(show['my_progress']/12)+1) * \
                        12  # Round up to the next cour

                if row in self.eps:
                    return (show['my_progress'], total, self.eps[row][0], self.eps[row][1])
                else:
                    return (show['my_progress'], total, None, None)
            elif column == ShowListModel.COL_NEXT_EP:
                return self.next_ep.get(row, '-')
            elif column == ShowListModel.COL_START_DATE:
                return self._date(show['start_date'])
            elif column == ShowListModel.COL_END_DATE:
                return self._date(show['end_date'])
            elif column == ShowListModel.COL_MY_START:
                return self._date(show['my_start_date'])
            elif column == ShowListModel.COL_LAST_UPDATED:
                return utils.format_local_time(show.get('my_last_update'))
            elif column == ShowListModel.COL_MY_FINISH:
                return self._date(show['my_finish_date'])
            elif column == ShowListModel.COL_MY_TAGS:
                return show.get('my_tags', '-')
            elif column == ShowListModel.COL_MY_STATUS:
                return self.mediainfo['statuses_dict'][show['my_status']]
        elif role == QtCore.Qt.ItemDataRole.BackgroundRole:
            return self.colors.get(row)
        elif role == QtCore.Qt.ItemDataRole.DecorationRole:
            if column == ShowListModel.COL_TITLE and show['id'] in self.playing:
                return getIcon('media-playback-start')
        elif role == QtCore.Qt.ItemDataRole.TextAlignmentRole:
            if column in [ShowListModel.COL_MY_PROGRESS, ShowListModel.COL_MY_SCORE]:
                return QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignVCenter
        elif role == QtCore.Qt.ItemDataRole.ToolTipRole:
            if column == ShowListModel.COL_PERCENT:
                tooltip = "Watched: %d<br>" % show['my_progress']
                if self.eps.get(row):
                    (aired_eps, library_eps) = self.eps.get(row)
                    if aired_eps:
                        tooltip += "Aired (estimated): %d<br>" % aired_eps
                    if library_eps:
                        tooltip += "Latest available: %d<br>" % max(
                            library_eps)
                tooltip += "Total: %d" % show['total']

                return tooltip
            elif column == ShowListModel.COL_LAST_UPDATED:
                return utils.format_local_time(show.get('my_last_update'))
        elif role == QtCore.Qt.ItemDataRole.EditRole:
            if column == ShowListModel.COL_MY_PROGRESS:
                return (show['my_progress'], show['total'], 0, 1)
            elif column == ShowListModel.COL_MY_SCORE:
                if isinstance(self.mediainfo['score_step'], float):
                    decimals = len(
                        str(self.mediainfo['score_step']).split('.')[1])
                else:
                    decimals = 0

                return (show['my_score'], self.mediainfo['score_max'], decimals, self.mediainfo['score_step'])
        elif role == QtCore.Qt.ItemDataRole.UserRole:
            if column == ShowListModel.COL_LAST_UPDATED:
                dt = show.get('my_last_update')
                return dt.timestamp() if dt is not None else 0

    def flags(self, index):
        if index.column() in self.editable_columns:
            return self.common_flags | QtCore.Qt.ItemFlag.ItemIsEditable
        else:
            return self.common_flags


class AddTableModel(QtCore.QAbstractTableModel):
    columns = ["Name", "Type", "Total"]

    def __init__(self, parent=None):
        self.results = None

        super().__init__(parent)

    def setResults(self, new_results):
        self.beginResetModel()
        self.results = new_results
        self.endResetModel()

    def rowCount(self, parent):
        if self.results:
            return len(self.results)
        else:
            return 0

    def columnCount(self, parent):
        return 3

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.ItemDataRole.DisplayRole and orientation == QtCore.Qt.Orientation.Horizontal:
            return self.columns[section]

    def data(self, index, role):
        row, column = index.row(), index.column()

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            item = self.results[row]

            if column == 0:
                return item.get('title')
            elif column == 1:
                return str(item.get('type', '?'))
            elif column == 2:
                return item.get('total', '?')


class AddListModel(QtCore.QAbstractListModel):
    """
    List model meant to be used with the Add show list view.

    It manages thumbnails and queues their downloads with the
    ThumbManager as necessary.
    """

    def __init__(self, parent=None, api_info=None):
        self.results = None
        self.thumbs = {}
        self.api_info = api_info

        self.pool = ThumbManager()
        self.pool.itemFinished.connect(self.gotThumb)

        super().__init__(parent)

    def gotThumb(self, iid, thumb):
        iid = int(iid)
        self.thumbs[iid] = thumb.scaled(
            100, 140, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)

        self.dataChanged.emit(self.index(iid), self.index(iid))

    def setResults(self, new_results):
        """ This method will process a new list of shows and get their
        thumbnails if necessary. """

        self.beginResetModel()

        self.results = new_results

        self.thumbs.clear()

        if self.results:
            for row, item in enumerate(self.results):
                if item.get('image'):
                    utils.make_dir(utils.to_cache_path())
                    filename = utils.to_cache_path("%s_%s_f_%s.jpg" % (
                        self.api_info['shortname'], self.api_info['mediatype'], item['id']))

                    if self.pool.exists(filename):
                        self.thumbs[row] = self.pool.getThumb(filename).scaled(
                            100, 140, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)
                    else:
                        self.pool.queueDownload(row, item['image'], filename)

        self.endResetModel()

    def rowCount(self, parent):
        if self.results:
            return len(self.results)
        else:
            return 0

    def data(self, index, role):
        row = index.row()
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            return self.results[row]
        elif role == QtCore.Qt.ItemDataRole.DecorationRole:
            return self.thumbs.get(row)
        elif role == QtCore.Qt.ItemDataRole.BackgroundRole:
            t = self.results[row].get('type')
            if t == utils.Type.TV:
                return QtGui.QColor(202, 253, 150)
            elif t == utils.Type.MOVIE:
                return QtGui.QColor(150, 202, 253)
            elif t == utils.Type.OVA or t == utils.Type.ONA:
                return QtGui.QColor(253, 253, 150)
            elif t == utils.Type.SP:
                return QtGui.QColor(253, 150, 150)
            else:
                return QtGui.QColor(250, 250, 250)

        return None


class AddListProxy(QtCore.QSortFilterProxyModel):
    def lessThan(self, left, right):
        leftData = self.sourceModel().data(left, QtCore.Qt.ItemDataRole.DisplayRole)
        rightData = self.sourceModel().data(right, QtCore.Qt.ItemDataRole.DisplayRole)

        return leftData['type'] < rightData['type']


class ShowListProxy(QtCore.QSortFilterProxyModel):
    filter_columns = None
    filter_status = None

    def setFilterStatus(self, status):
        self.filter_status = status
        self.invalidateFilter()

    def clearColumnFilters(self):
        self.filters = {}

    def setFilterColumns(self, columns):
        self.filter_columns = columns
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        if self.filter_status is not None and self.sourceModel().showlist[source_row]['my_status'] != self.filter_status:
            return False

        if self.filter_columns:
            for col in range(self.sourceModel().columnCount(source_parent)):
                index = self.sourceModel().index(source_row, col)
                if (col in self.filter_columns and
                        self.filter_columns[col] not in str(self.sourceModel().data(index, QtCore.Qt.ItemDataRole.DisplayRole))):
                    return False

        return super(ShowListProxy, self).filterAcceptsRow(source_row, source_parent)

    def lessThan(self, left, right):
        col = left.column()

        if col == ShowListModel.COL_LAST_UPDATED:
            lv = self.sourceModel().data(left, QtCore.Qt.ItemDataRole.UserRole)
            rv = self.sourceModel().data(right, QtCore.Qt.ItemDataRole.UserRole)

            lnum = lv if isinstance(lv, (int, float)) else 0
            rnum = rv if isinstance(rv, (int, float)) else 0

            return int(lnum) < int(rnum)

        return super().lessThan(left, right)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ntpath
import time
import urllib.error
import urllib.parse
import urllib.request
import xml.dom.minidom as xdmd

import trackma.utils as utils
from trackma.tracker import tracker

NOT_RUNNING = 0
ACTIVE = 1
CLAIMED = 2
PLAYING = 3
PAUSED = 4
IDLE = 5


class PlexTracker(tracker.TrackerBase):
    name = 'Tracker (Plex)'

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        self.config = config

        self.host_port = self.config['plex_host']+":"+self.config['plex_port']
        self.status_log = [None, None]
        self.token = self._get_plex_token()
        self.http_protocol = 'https://' if self.config['plex_ssl'] else 'http://'
        super().__init__(messenger, tracker_list, config, watch_dirs, redirections)

    def get_plex_status(self):
        # returns the plex status of the first active session
        try:
            active = int(self._get_sessions_info("MediaContainer", "size"))

            if active:
                return ACTIVE
            else:
                return IDLE
        except urllib.error.URLError as e:
            if e.code == 401:
                return CLAIMED
            else:
                return NOT_RUNNING

    def playing_file(self):
        # returns the filename of the currently playing file
        if self.get_plex_status() == IDLE:
            return None

        meta = self._get_sessions_info("Video", "key")
        meta_url = self.http_protocol+self.host_port+meta
        mres = self._get_xml_info(meta_url, "Part", "file")
        name = urllib.parse.unquote(ntpath.basename(mres))
        xstate = self._get_sessions_info("Player", "state")

        if xstate == "playing":
            state = PLAYING
        else:
            state = PAUSED

        return (name, state)

    def timer_from_file(self):
        # returns 80% of video duration for the update timer,
        # roughly the time of the video minus the OP and ED
        if self.get_plex_status() == IDLE:
            return None

        try:
            duration = int(self._get_sessions_info("Video", "duration"))
            return round((duration*0.80)/60000)*60
        except IndexError:
            return None

    def observe(self, config, watch_dirs):
        self.msg.info("Using Plex.")

        while self.active:
            self.status_log.append(self.get_plex_status())

            if self.status_log[-1] == ACTIVE or self.status_log[-1] == IDLE:
                if self.status_log[-1] == IDLE and self.status_log[-2] == NOT_RUNNING:
                    self.msg.info("Reconnected to Plex.")

                if self.config['plex_obey_update_wait_s']:
                    self.wait_s = config['tracker_update_wait_s']
                else:
                    self.wait_s = self.timer_from_file()

                try:
                    xuser = self._get_sessions_info("User", "title")

                    player = self.playing_file()
                    (state, show_tuple) = self._get_playing_show(player[0])

                    if self.token:
                        if self.config['plex_user'] == xuser:
                            self.view_offset = int(self._get_sessions_info("Video", "viewOffset"))
                            self.update_show_if_needed(state, show_tuple)

                            if player[1] == PAUSED:
                                self.pause_timer()
                            elif player[1] == PLAYING:
                                self.resume_timer()
                    else:
                        self.update_show_if_needed(state, show_tuple)

                        if player[1] == PAUSED:
                            self.pause_timer()
                        elif player[1] == PLAYING:
                            self.resume_timer()
                except IndexError:
                    if self.status_log[-1] == IDLE:
                        self.last_filename = None
                        self.update_show_if_needed(0, None)
                    else:
                        pass
            elif self.status_log[-1] == CLAIMED and self.status_log[-2] == CLAIMED:
                self.msg.warn("Claimed Plex Media Server, login in the settings and restart trackma.")
            elif self.status_log[-1] == NOT_RUNNING and self.status_log[-2] == NOT_RUNNING:
                self.msg.warn("Plex Media Server is not running.")

            del self.status_log[0]

            # Wait for the interval before running check again
            time.sleep(config['tracker_interval'])

    def _get_plex_token(self):
        username = urllib.parse.quote(self.config['plex_user'])
        password = urllib.parse.quote(self.config['plex_passwd'])
        uuid = self.config['plex_uuid']

        if not (username and password):
            return ''

        body = bytes('user[login]=%s&user[password]=%s' %
                     (username, password), "utf-8")
        headers = {'X-Plex-Client-Identifier': uuid,
                   'X-Plex-Product': "Trackma",
                   'X-Plex-Version': utils.VERSION}

        req = urllib.request.Request(
            'https://plex.tv/users/sign_in.xml', body, headers=headers)
        response = urllib.request.urlopen(req)
        data = response.read().decode("utf-8")

        tdoc = xdmd.parseString(data)
        token = tdoc.getElementsByTagName("user")[0].getAttribute("authToken")

        return "?X-Plex-Token="+token

    def _get_xml_info(self, url, tag, attr):
        try:
            uop = urllib.request.urlopen(url)
        except urllib.error.URLError:
            uop = urllib.request.urlopen(url+self.token)

        doc = xdmd.parse(uop)
        elem = doc.getElementsByTagName(tag)
        res = elem[0].getAttribute(attr)

        # if there's more than one session and there's a token lookup the matching user
        if len(elem) > 1 and self.token:
            for e in elem:
                etag = e.getElementsByTagName("User")
                xuser = etag[0].getAttribute(
                    "title") if etag else e.getAttribute("title")

                if xuser == self.config['plex_user']:
                    res = e.getAttribute(attr)
                    break

        return res

    def _get_sessions_info(self, tag, attr):
        # Get the required info from the /status/sessions url
        session_url = self.http_protocol+self.host_port+"/status/sessions"
        info = self._get_xml_info(session_url, tag, attr)

        return info



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import threading
import time

from trackma import utils
from trackma.parser import get_parser_class
from trackma.messenger import Messenger


class TrackerBase(object):
    msg: Messenger
    active = True
    list = None
    last_show_tuple = None
    last_filename = None
    last_state = utils.Tracker.NOVIDEO
    last_time = 0
    last_updated = False
    last_close_queue = None
    timer = None

    name = 'Tracker'

    signals = {
        'state': None,
        'detected': None,
        'playing': None,
        'removed': None,
        'update': None,
        'unrecognised': None,
    }

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        self.msg = messenger.with_classname(self.name)
        self.msg.info('Initializing...')

        self.list = tracker_list
        self.config = config
        self.redirections = redirections
        # Reverse sorting for prefix matching
        self.watch_dirs = tuple(sorted(watch_dirs, reverse=True))
        self.wait_s = None

        self.timer = None
        self.timer_paused = None
        self.timer_offset = 0
        self.parser_class = get_parser_class(self.msg, self.config['title_parser'])

        self.view_offset = None

        tracker_args = (config, watch_dirs)
        tracker_t = threading.Thread(target=self.observe, args=tracker_args)
        tracker_t.daemon = True

        self.msg.debug('Enabling tracker...')
        tracker_t.start()

    def set_message_handler(self, message_handler):
        """Changes the message handler function on the fly."""
        self.msg = message_handler.with_classname(self.name)

    def disable(self):
        self.msg.info('Unloading...')
        self.active = False

    def update_list(self, tracker_list):
        self.list = tracker_list

    def connect_signal(self, signal, callback):
        try:
            self.signals[signal] = callback
        except KeyError:
            raise utils.EngineFatal("Invalid signal.")

    def observe(self, config, watch_dirs):
        raise NotImplementedError

    def get_status(self):
        return {
            'state': self.last_state,
            'timer': self.timer,
            'viewOffset': self.view_offset,
            'paused': bool(self.timer_paused),
            'show': self.last_show_tuple,
            'filename': self.last_filename,
        }

    def _emit_signal(self, signal, *args):
        try:
            if self.signals[signal]:
                self.signals[signal](*args)
        except KeyError:
            raise Exception("Call to undefined signal.")

    def update_timer(self, state, show_tuple):
        if self.timer_paused:
            return

        (show, episode) = show_tuple

        self.timer = int(
            1
            + (self.wait_s or self.config['tracker_update_wait_s'])
            + self.timer_offset
            - (time.time() - self.last_time)
        )
        self._emit_signal('state', self.get_status())

        if self.timer <= 0:
            # Perform show update
            self.last_updated = True

            def action(state=state):
                if state == utils.Tracker.PLAYING:
                    return self._emit_signal('update', show, episode)
                elif state == utils.Tracker.NOT_FOUND:
                    return self._emit_signal('unrecognised', show, episode)

            if self.config['tracker_update_close']:
                self.msg.info('Waiting for the player to close.')
                self.last_close_queue = action
            else:
                action()

    def _ignore_current(self):
        # Stops attempt to update current episode
        self.last_updated = True
        self.last_state = utils.Tracker.IGNORED
        self.timer = None
        self._emit_signal('state', self.get_status())

    def _update_state(self, state):
        # Call when show or state is changed. Perform queued update if any.
        if self.last_close_queue:
            self.last_close_queue()
            self.last_close_queue = None

        # Clear up pause and set our new time offset
        self.timer_paused = None
        self.timer_offset = 0
        self.last_time = time.time()

        # Emit the new playing signal
        if self.last_show_tuple:
            (last_show, last_show_ep) = self.last_show_tuple
            if last_show['id']:
                self._emit_signal(
                    'playing', last_show['id'], False, last_show_ep)

    def pause_timer(self):
        if not self.timer_paused:
            self.timer_paused = time.time()

            self._emit_signal('state', self.get_status())

    def resume_timer(self):
        if self.timer_paused:
            self.timer_offset += time.time() - self.timer_paused
            self.timer_paused = None

            self._emit_signal('state', self.get_status())

    def update_show_if_needed(self, state, show_tuple):
        # If the state and show are unchanged, skip to countdown
        if show_tuple and state == self.last_state and show_tuple == self.last_show_tuple and not self.last_updated:
            self.update_timer(state, show_tuple)
            return

        if show_tuple and show_tuple != self.last_show_tuple:
            (show, episode) = show_tuple
            self._update_state(state)
            # There's a new show/ep detected, so let's save the show information
            self.last_show_tuple = show_tuple
            self.last_updated = False
            if state == utils.Tracker.PLAYING:
                self._emit_signal('playing', show['id'], True, episode)
                # Check if we shouldn't update the show
                expected_next_ep = show['my_progress'] + 1
                if self.config['tracker_ignore_not_next'] and episode != expected_next_ep:
                    self.msg.warn(
                        'Not playing the next episode of {} (expected: {}, found: {}). Ignoring.'
                            .format(show['title'], expected_next_ep, episode),
                    )
                    self._ignore_current()
                    return
                if episode == show['my_progress']:
                    self.msg.warn('Playing the current episode of %s. Ignoring.' % show['title'])
                    self._ignore_current()
                    return
                if episode < 1 or (show['total'] and episode > show['total']):
                    self.msg.warn('Playing an invalid episode of %s. Ignoring.' % show['title'])
                    self._ignore_current()
                    return

            # Start our countdown
            (show, episode) = show_tuple
            if state == utils.Tracker.PLAYING:
                self.msg.info('Will update %s - %d' % (show['title'], episode))
            elif state == utils.Tracker.NOT_FOUND:
                self.msg.info('Will add %s - %d' % (show['title'], episode))

            self.update_timer(state, show_tuple)
        elif self.last_state != state:
            self._update_state(state)

            # React depending on state
            if state == utils.Tracker.NOVIDEO:  # No video is playing
                # Video didn't get to update phase before it was closed
                if self.last_state == utils.Tracker.PLAYING and not self.last_updated:
                    self.msg.info('Player was closed before update.')
            # There's a new video playing but the regex didn't recognize the format
            elif state == utils.Tracker.UNRECOGNIZED:
                self.msg.warn('Found video but the file name format couldn\'t be recognized.')
            elif state == utils.Tracker.NOT_FOUND:  # There's a new video playing but an associated show wasn't found
                self.msg.warn('Found player but show not in list.')

            self.last_show_tuple = None
            self.last_updated = False
            self.timer = None

        self.last_state = state
        self._emit_signal('state', self.get_status())

    def _get_playing_show(self, filename):
        if not self.active:
            # Don't do anything if the Tracker is disabled
            return (utils.Tracker.NOVIDEO, None)

        if filename:
            if filename == self.last_filename:
                # It's the exact same filename, there's no need to do the processing again
                return (self.last_state, self.last_show_tuple)

            self.last_filename = filename
            self.msg.debug("Guessing filename: {}".format(filename))

            # Trim out watch dir
            if os.path.isabs(filename):
                for watch_prefix in self.watch_dirs:
                    if filename.startswith(watch_prefix):
                        filename = filename[len(watch_prefix):].lstrip(os.path.sep)
                        break

            # Invoke the parser to extract show title and episode.
            aie = self.parser_class(self.msg, filename)
            (show_title, show_ep) = (aie.getName(), aie.getEpisode())
            if not show_title:
                # Format not recognized
                return (utils.Tracker.UNRECOGNIZED, None)

            playing_show = utils.guess_show(show_title, self.list)
            self.msg.debug("Show guess: {}: {} - {}".format(show_title, playing_show, show_ep))

            if playing_show:
                (redirected_show, redirected_ep) = utils.redirect_show(
                    (playing_show, show_ep), self.redirections, self.list)
                if (redirected_show, redirected_ep) != (playing_show, show_ep):
                    self.msg.debug("Redirected to: {} - {}".format(redirected_show, redirected_ep))
                    (playing_show, show_ep) = (redirected_show, redirected_ep)

                return (utils.Tracker.PLAYING, (playing_show, show_ep))
            else:
                # Show not in list
                if self.config['tracker_not_found_prompt']:
                    # Dummy show to search for
                    show = {'id': 0, 'title': show_title}
                    return (utils.Tracker.NOT_FOUND, (show, show_ep))
                else:
                    return (utils.Tracker.NOT_FOUND, None)
        else:
            self.last_filename = None
            return (utils.Tracker.NOVIDEO, None)  # Not playing



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from gi import require_version

require_version('Gtk', '3.0')
require_version('Gdk', '3.0')

from trackma import utils  # noqa: E402
from trackma.ui.gtk.window import TrackmaWindow  # noqa: E402
from gi.repository import GLib, Gio, Gtk  # noqa: E402


class TrackmaApplication(Gtk.Application):
    __gtype_name__ = 'TrackmaApplication'

    def __init__(self):
        super().__init__(
            application_id="com.github.z411.TrackmaGtk",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.NON_UNIQUE
        )

        self.debug = False
        self.window = None
        self.add_main_option(
            "debug", ord(
                "d"), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show debugging information", None
        )

    def do_startup(self):
        Gtk.Application.do_startup(self)

        self._register_accelerators()

        action = Gio.SimpleAction.new('quit', None)
        action.connect("activate", self._on_quit)
        self.add_action(action)

    def do_activate(self):
        try:
            self.create_window()
        except utils.TrackmaFatal as e:
            self.message_error(e)

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        options = options.end().unpack()
        self.debug = "debug" in options
        self.activate()
        return 0

    def _register_accelerators(self):
        def set_accel(name, accel):
            accel = accel if isinstance(accel, tuple) else (accel,)
            self.set_accels_for_action(name, accel)

        accels = (
            ('win.search', '<Primary>A'),
            ('win.synchronize', '<Primary>S'),
            ('win.upload', '<Primary>E'),
            ('win.download', '<Primary>D'),
            ('win.scanfiles', '<Primary>L'),
            ('win.show-help-overlay', '<Primary>question'),
            ('app.quit', '<Primary>Q'),

            # Shows
            ('win.play_next', '<Primary>N'),
            ('win.play_random', '<Primary>R'),
            ('win.episode_add', '<Primary>Right'),
            ('win.episode_remove', '<Primary>Left'),
            ('win.delete', ('Delete', 'KP_Delete')),
            ('win.copy', '<Primary>C')
        )

        for (name, accel) in accels:
            set_accel(name, accel)

    def create_window(self):
        if not self.window:
            self.window = TrackmaWindow(self, self.debug)

        if not self.window._engine:
            self.window.init_account_selection()

    @staticmethod
    def message_error(error):
        md = Gtk.MessageDialog(
            None,
            Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.ERROR,
            Gtk.ButtonsType.CLOSE,
            str(error)
        )
        md.run()
        md.destroy()

    def _on_quit(self, action, param):
        self.window._quit()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
import sys
import threading
import time

from trackma import utils


class Data:
    """
    Data Handler Class

    Class for keeping data in memory, handling list cache and
    update queues. This module cares about keeping the data
    safe and up to date, and it handles commands given by the engine,
    and communicates with the API whenever it's necessary.

    messenger: Messenger object to send useful messages to
    config: Parsed configuration dictionary (given by the engine)

    """
    name = 'Data'
    # 8: Added "ONA" anime type
    version = 8

    msg = None
    api = None
    showlist = None
    infocache = dict()
    queue = list()
    config = dict()
    meta = {'lastget': 0, 'lastsend': 0, 'version': '', 'apiversion': '',
            'altnames': {}, 'library': {}, 'library_cache': {}, }

    autosend_timer = None

    signals = {
        'show_synced':       None,
        'sync_complete':     None,
        'queue_changed':     None,
    }

    def __init__(self, messenger, config, account, mediatype):
        """Checks if the config is correct and creates an API object."""
        self.msg = messenger.with_classname(self.name)
        self.config = config
        self.msg.info("Initializing...")

        # Get filenames
        userfolder = "%s.%s" % (account['username'], account['api'])
        self.userconfig_file = utils.to_data_path(userfolder, 'user.json')

        # Handle userconfig and media type to load
        self._load_userconfig()
        if mediatype:
            self.userconfig['mediatype'] = mediatype
            self._save_userconfig()

        # Import the API
        libbase = account['api']
        libname = "lib{0}".format(libbase)
        try:
            modulename = "trackma.lib.{0}".format(libname)
            __import__(modulename)
            apimodule = sys.modules[modulename]
        except ImportError as e:
            raise utils.DataFatal("Couldn't import API module: %s" % e)

        # Instance API
        libclass = getattr(apimodule, libname)
        self.api = libclass(self.msg, account, self.userconfig)

        # Get API version
        self.api_version = self.api.api_info['version']

        # Set mediatype
        mediatype = self.userconfig.get('mediatype')
        self.msg.info("Using %s (%s)" % (libname, mediatype))

        # Get filenames
        self.queue_file = utils.to_data_path(
            userfolder, '%s.queue' % mediatype)
        self.info_file = utils.to_data_path(userfolder,  '%s.info' % mediatype)
        self.cache_file = utils.to_data_path(userfolder, '%s.list' % mediatype)
        self.meta_file = utils.to_data_path(userfolder, '%s.meta' % mediatype)
        self.lock_file = utils.to_data_path(userfolder,  'lock')

        # Connect signals
        self.api.connect_signal('show_info_changed', self.info_update)
        self.api.connect_signal('userconfig_changed', self.userconfig_update)

    def _emit_signal(self, signal, *args):
        if self.signals[signal]:
            self.signals[signal](*args)

    def _is_queue_ready(self):
        # Checks if queue should be sent ASAP
        # Note: Hours setting is DEPRECATED!
        return (self.config['autosend'] == 'always' or
                (self.config['autosend'] == 'hours' and time.time() - self.meta['lastsend'] >= self.config['autosend_hours']*3600) or
                (self.config['autosend'] == 'minutes' and time.time() - self.meta['lastsend'] >= self.config['autosend_minutes']*60) or
                (self.config['autosend'] == 'size' and len(self.queue) >= self.config['autosend_size']))

    def connect_signal(self, signal, callback):
        try:
            self.signals[signal] = callback
        except KeyError:
            raise utils.DataFatal("Invalid signal.")

    def set_message_handler(self, message_handler):
        self.msg = message_handler.with_classname(self.name)
        self.api.set_message_handler(self.msg)

    def start(self):
        """
        Does all necessary tasks to start the data handler

        This should be called before doing any other operation with the data handler,
        as it loads the list cache (or downloads it if necessary) and queue.

        """
        # Lock the database
        self.msg.debug("Locking database...")
        self._lock()

        # Load different caches
        if self._meta_exists():
            self._load_meta()

        if self._queue_exists() and self.meta.get('version') == self.version and self.meta.get('apiversion') == self.api_version:
            self._load_queue()
            self._emit_signal('queue_changed', self.queue)

        if self._info_exists() and self.meta.get('version') == self.version and self.meta.get('apiversion') == self.api_version:
            # Load info cache only if we're on the same database version
            self._load_info()

        # If there is a list cache, load from it
        # otherwise query the API for a remote list
        if self._cache_exists() and self.meta.get('version') == self.version and self.meta.get('apiversion') == self.api_version:
            # Auto-send: Process the queue if we're beyond the auto-send time limit for some reason
            if self._is_queue_ready():
                self.process_queue()

            # Auto-retrieve: Redownload list if any autoretrieve condition is met
            if (self.config['autoretrieve'] == 'always' or
                (self.config['autoretrieve'] == 'days' and
                 time.time() - self.meta['lastget'] > self.config['autoretrieve_days'] * 84600) or
                    self.meta.get('version') != self.version):
                try:
                    # Make sure we process the queue first before overwriting the list
                    # We don't want users losing their changes
                    self.process_queue()
                    self.download_data()
                except utils.APIError as e:
                    self.msg.warn("Couldn't download list! Using cache.")
                    self._load_cache()
            elif not self.showlist:
                # If the cache wasn't loaded before, do it now
                self._load_cache()
        else:
            try:
                self.download_data()
            except utils.APIError as e:
                raise utils.APIFatal(str(e))

        # Create autosend thread if needed
        # Note: Hours setting is DEPRECATED!
        if self.config['autosend'] in ('minutes', 'hours'):
            self.autosend()

        return (self.api.api_info, self.api.media_info())

    def unload(self, force=False):
        """
        Does unloading of the data handler

        This should be called whenever the data handler won't be used anymore,
        as it does necessary operations to close the API and the data handler itself.

        """
        self.msg.debug("Unloading...")

        # Cancel autosend thread
        if self.autosend_timer:
            self.autosend_timer.cancel()

        # We push changes if specified on config file
        if not force:
            if self.config['autosend_at_exit']:
                self.process_queue()

            self._save_meta()

        self._unlock()

    def get(self):
        """Get list from memory"""
        return self.showlist

    def search(self, criteria, method):
        # Tell API to search
        results = self.api.search(criteria, method)
        self.api.logout()
        if results:
            return results

        raise utils.DataError('No results.')

    def queue_add(self, show):
        """
        Queues a show add

        Calls this to add a show to the list, and the remote add
        will be queued.

        show: Show dictionary

        """
        showid = show['id']

        # Add to the list
        if self.showlist.get(showid):
            raise utils.DataError("Show already in the list.")

        self.showlist[showid] = show

        # Check if the show add is already in queue
        exists = False
        for q in self.queue:
            if q['id'] == showid and q['action'] == 'add':
                # This shouldn't happen
                raise utils.DataError("Show already in the queue.")

        if not exists:
            # Use the whole show as a queue item
            item = show
            item['action'] = 'add'
            self.queue.append(item)

        show['queued'] = True

        self._save_queue()
        self._save_cache()
        self._emit_signal('queue_changed', self.queue)
        self.msg.info("Queued add for %s" % show['title'])

    def queue_update(self, show, key, value):
        """
        Queues a show update

        Call this to change anything of an item in the list, as it will be
        modified locally and queued for update in the next remote sync.

        show: Show dictionary
        key: The key that will be modified (it must exist beforehand)
        value: The value that it should be changed to
        """
        if key not in show:
            raise utils.DataError('Invalid key for queue update.')

        # Do update on memory
        show[key] = value

        # Check if the show update is already in queue
        exists = False
        for q in self.queue:
            if q['id'] == show['id'] and q['action'] in ['add', 'update']:
                # Add the changed value to the already existing queue item
                q[key] = value
                exists = True
                break

        if not exists:
            # Create queue item and append it
            item = {'id': show['id'],
                    'my_id': show['my_id'],
                    'action': 'update',
                    'title': show['title'],
                    }
            item[key] = value
            self.queue.append(item)

        show['queued'] = True

        self._save_queue()
        self._save_cache()
        self._emit_signal('queue_changed', self.queue)
        self.msg.info("Queued update for %s" % show['title'])
        self.msg.debug("Queued: {} -> {}".format(key, value))

        # Immediately process the action if necessary
        if self._is_queue_ready():
            self.process_queue()

    def queue_delete(self, show):
        """
        Queues a show delete

        Calls this to delete a show from the list, and the remote delete
        will be queued.

        show: Show dictionary

        """
        showid = show['id']

        # Delete from the list
        if not self.showlist.get(showid):
            raise utils.DataError("Show not in the list.")

        item = self.showlist.pop(showid)

        # Check if the show add is already in queue
        exists = False
        for q in self.queue:
            if q['id'] == showid and q['action'] == 'delete':
                # This shouldn't happen
                raise utils.DataError("Show delete already in the queue.")

        if not exists:
            # Use the whole show as a queue item
            item['action'] = 'delete'
            self.queue.append(item)

        show['queued'] = True

        self._save_queue()
        self._save_cache()
        self._emit_signal('queue_changed', self.queue)
        self.msg.info("Queued delete for %s" % item['title'])

    def queue_clear(self):
        """Clears the queue completely."""
        if self.queue:
            self.queue = []
            self._save_queue()
            self._emit_signal('queue_changed', self.queue)
            self.msg.info("Cleared queue.")

    def process_queue(self):
        """
        Send updates in queue to the API

        It starts sending all the queued updates to the API for it to update
        the remote list. Any successful updates get removed from the queue,
        and failed updates stay there to be processed the next time.

        """
        if self.queue:
            self.msg.info('Processing queue...')

            # Load the cache if it wasn't loaded for some reason
            if not self.showlist:
                self._load_cache()

            # Check log-in TODO
            # try:
            #    self.api.check_credentials()
            # except utils.APIError as e:
            #    raise utils.DataError("Can't process queue, will leave unsynced. Reason: %s" % e)

            # Run through queue
            items_processed = []
            items_failed = []
            while True:
                try:
                    item = self.queue.pop(0)
                except IndexError:
                    break

                showid = item['id']

                try:
                    show = self.showlist[showid]
                except KeyError:
                    show = None

                try:
                    # Call the API to do the requested operation
                    operation = item.get('action')
                    if operation == 'add':
                        my_id = self.api.add_show(item)

                        if my_id:
                            show['my_id'] = my_id
                    elif operation == 'update':
                        my_last_update = self.api.update_show(item)
                        
                        if my_last_update:
                            show['my_last_update'] = my_last_update
                    elif operation == 'delete':
                        self.api.delete_show(item)
                    else:
                        self.msg.warn("Unknown operation in queue (%s), skipping..." % repr(operation))

                    if self.showlist.get(showid):
                        self.showlist[showid]['queued'] = False
                        self._emit_signal('show_synced', show, item)

                    items_processed.append((show, item))
                    self._emit_signal('queue_changed', self.queue)
                except utils.APIError as e:
                    self.msg.warn("Can't process %s, will leave unsynced." % item['title'])
                    self.msg.debug("Info: %s" % e)
                    items_failed.append(item)
                except NotImplementedError:
                    self.msg.warn("Operation not implemented in API. Skipping...")
                    items_failed.append(item)
                # except TypeError:
                #    self.msg.warn("%s not in list, unexpected. Not changing queued status." % showid)

            if items_failed:
                self.queue += items_failed

            self.api.logout()
            self._save_cache()
            self._save_queue()
            self._emit_signal('sync_complete', items_processed)
        else:
            self.msg.debug('No items in queue.')

        self.meta['lastsend'] = time.time()

    def info_get(self, show):
        try:
            showid = show['id']
            return self.infocache[showid]
        except KeyError:
            return self.api.request_info([show])[0]

    def info_update(self, shows):
        for show in shows:
            showid = show['id']
            self.infocache[showid] = show

        self._save_info()

    def userconfig_update(self):
        self._save_userconfig()

    def altname_get(self, showid):
        return self.meta['altnames'].get(showid, '')

    def altname_set(self, showid, altname):
        self.meta['altnames'][showid] = altname

    def altname_clear(self, showid):
        if showid in self.meta['altnames']:
            del self.meta['altnames'][showid]

    def altnames_get(self):
        return self.meta['altnames']

    def library_get(self):
        return self.meta['library']

    def library_save(self, library):
        self.meta['library'] = library

    def library_cache_get(self):
        return self.meta['library_cache']

    def library_cache_save(self, library_cache):
        self.meta['library_cache'] = library_cache

    def get_show_attr(self, show, key):
        return show.get(key)

    def set_show_attr(self, show, key, value):
        show[key] = value

    def get_show_titles(self, show):
        return [show['title']] + show['aliases']

    def get_altnames_map(self):
        return {name.lower(): showid for showid, name in self.altnames_get().items()}

    def autosend(self):
        # Check if we should autosend now
        if self._is_queue_ready():
            self.process_queue()

        # Repeat check only if the settings are still on autosend
        # Note: Hours setting is DEPRECATED!
        if self.config['autosend'] in ('minutes', 'hours'):
            self.autosend_timer = threading.Timer(
                3600 if self.config['autosend'] == 'hours' else 60, self.autosend)
            self.autosend_timer.daemon = True
            self.autosend_timer.start()

    def _load_cache(self):
        self.msg.debug("Reading cache...")
        self.showlist = utils.load_data(self.cache_file)

    def _save_cache(self):
        self.msg.debug("Saving cache...")
        utils.save_data(self.showlist, self.cache_file)

    def _load_info(self):
        self.msg.debug("Reading info DB...")
        self.infocache = utils.load_data(self.info_file)

    def _save_info(self):
        self.msg.debug("Saving info DB...")
        utils.save_data(self.infocache, self.info_file)

    def _load_userconfig(self):
        self.msg.debug("Reading userconfig...")
        self.userconfig = utils.parse_config(
            self.userconfig_file, utils.userconfig_defaults)

    def _save_userconfig(self):
        self.msg.debug("Saving userconfig...")
        utils.save_config(self.userconfig, self.userconfig_file)

    def _load_queue(self):
        self.msg.debug("Reading queue...")
        self.queue = utils.load_data(self.queue_file)

    def _save_queue(self):
        self.msg.debug("Saving queue...")
        utils.save_data(self.queue, self.queue_file)

    def _load_meta(self):
        self.msg.debug("Reading metadata...")
        loadedmeta = utils.load_data(self.meta_file)
        self.meta.update(loadedmeta)

    def _save_meta(self):
        self.msg.debug("Saving metadata...")
        utils.save_data(self.meta, self.meta_file)

    def download_data(self):
        """Downloads the remote list and overwrites the cache"""
        self.showlist = self.api.fetch_list()

        if self.api.api_info['merge']:
            # The API needs information to be merged from the
            # info database
            missing = []
            for show in self.showlist.values():
                # Here we search the information in the local
                # info database. If it isn't available, add it
                # to the missing list for them to be requested
                # to the API later.
                showid = show['id']

                try:
                    info = self.infocache[showid]
                except KeyError:
                    missing.append(show)
                    continue

                self.api.merge(show, info)

            # Here we request the missing items and merge them
            # immediately with the list.
            if missing:
                infos = self.api.request_info(missing)
                for info in infos:
                    showid = info['id']
                    self.api.merge(self.showlist[showid], info)

        self._save_cache()
        self.api.logout()

        # Update last retrieved time
        self.meta['lastget'] = time.time()
        self.meta['version'] = self.version
        self.meta['apiversion'] = self.api_version
        self._save_meta()

    def _cache_exists(self):
        return os.path.isfile(self.cache_file)

    def _info_exists(self):
        return os.path.isfile(self.info_file)

    def _queue_exists(self):
        return os.path.isfile(self.queue_file)

    def _meta_exists(self):
        return os.path.isfile(self.meta_file)

    def _lock(self):
        """Creates the database lock, returns an exception if it
        already exists"""
        if self.config['debug_disable_lock']:
            return

        if os.path.isfile(self.lock_file):
            raise utils.DataFatal("Database is locked by another process. "
                                  "If you\'re sure there's no other process is using it, "
                                  "remove the file ~/.trackma/lock")

        f = open(self.lock_file, 'w')
        f.close()

    def _unlock(self):
        """Removes the database lock"""
        if self.config['debug_disable_lock']:
            return

        os.unlink(self.lock_file)

    def get_api_info(self):
        return (self.api.api_info, self.api.media_info())



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt6 import QtCore, QtGui
from PyQt6.QtWidgets import (QAbstractItemView, QCheckBox, QComboBox, QDialog, QDialogButtonBox, QFormLayout,
                             QHBoxLayout, QHeaderView, QLabel, QLineEdit, QMessageBox, QPushButton, QTableWidget,
                             QTableWidgetItem, QVBoxLayout)

from trackma import utils


class AccountDialog(QDialog):

    selected = QtCore.pyqtSignal(int, bool)
    aborted = QtCore.pyqtSignal()
    selected_account_num = None

    def __init__(self, parent, accountman):
        QDialog.__init__(self, parent)

        self.accountman = accountman

        layout = QVBoxLayout()

        self.setWindowTitle('Select Account')

        # Create list
        self.table = QTableWidget()
        self.table.horizontalHeader().setHighlightSections(False)
        self.table.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.table.verticalHeader().hide()
        self.table.setGridStyle(QtCore.Qt.PenStyle.NoPen)
        self.table.doubleClicked.connect(self.select)
        self.table.itemSelectionChanged.connect(self.update_selection)

        bottom_layout = QHBoxLayout()
        self.remember_chk = QCheckBox('Remember')
        cancel_btn = QPushButton('Cancel')
        cancel_btn.clicked.connect(self.cancel)
        add_btn = QPushButton('Add')
        add_btn.clicked.connect(self.add)
        self.edit_btns = QComboBox()
        self.edit_btns.blockSignals(True)
        self.edit_btns.addItem('Edit...')
        self.edit_btns.addItem('Update')
        self.edit_btns.addItem('Delete')
        self.edit_btns.addItem('Purge')
        self.edit_btns.setItemData(
            1, 'Change the local password/PIN for this account', QtCore.Qt.ItemDataRole.ToolTipRole)
        self.edit_btns.setItemData(
            2, 'Remove this account from Trackma', QtCore.Qt.ItemDataRole.ToolTipRole)
        self.edit_btns.setItemData(
            3, 'Clear local DB for this account', QtCore.Qt.ItemDataRole.ToolTipRole)
        self.edit_btns.setCurrentIndex(0)
        self.edit_btns.blockSignals(False)
        self.edit_btns.activated.connect(self.s_edit)
        self.select_btn = QPushButton('Select')
        self.select_btn.clicked.connect(self.select)
        self.select_btn.setDefault(True)
        bottom_layout.addWidget(self.remember_chk)
        bottom_layout.addWidget(cancel_btn)
        bottom_layout.addWidget(add_btn)
        bottom_layout.addWidget(self.edit_btns)
        bottom_layout.addWidget(self.select_btn)

        # Get icons
        self.icons = dict()
        for libname, lib in utils.available_libs.items():
            self.icons[libname] = QtGui.QIcon(lib[1])

        # Populate list
        self.update()
        self.rebuild()

        # Finish layout
        layout.addWidget(self.table)
        layout.addLayout(bottom_layout)
        self.setLayout(layout)

    def update(self):
        self.remember_chk.setChecked(self.accountman.get_default() is not None)

    def update_selection(self):
        if self.table.selectedItems():
            self.selected_account_num = self.table.selectedItems()[0].num
        else:
            self.selected_account_num = None
        self.edit_btns.setDisabled(not self.selected_account_num)
        self.select_btn.setDisabled(not self.selected_account_num)

    def add(self):
        result = AccountAddDialog.do(icons=self.icons)
        if result:
            (username, password, api, extra) = result
            self.accountman.add_account(username, password, api, extra)
            self.rebuild()

    def s_edit(self, index):
        self.edit_btns.blockSignals(True)
        self.edit_btns.setCurrentIndex(0)
        self.edit_btns.blockSignals(False)

        if index == 1:
            self.edit()
        elif index == 2:
            self.delete()
        elif index == 3:
            self.purge()

    def edit(self):
        acct = self.accountman.get_account(self.selected_account_num)
        result = AccountAddDialog.do(icons=self.icons,
                                     edit=True,
                                     username=acct['username'],
                                     password=acct['password'],
                                     api=acct['api'])
        if result:
            (username, password, api, extra) = result
            self.accountman.edit_account(
                self.selected_account_num, username, password, api, extra)
            self.rebuild()

    def delete(self):
        reply = QMessageBox.question(
            self, 'Confirmation', 'Do you want to delete the selected account?', QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

        if reply == QMessageBox.StandardButton.Yes:
            self.accountman.delete_account(self.selected_account_num)
            self.rebuild()

    def purge(self):
        reply = QMessageBox.question(
            self, 'Confirmation', 'Do you want to purge the selected account\'s local data?', QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

        if reply == QMessageBox.StandardButton.Yes:
            self.accountman.purge_account(self.selected_account_num)
            self.rebuild()

    def rebuild(self):
        self.table.clear()

        columns = ['Username', 'Site']
        self.table.setColumnCount(len(columns))
        self.table.setHorizontalHeaderLabels(columns)
        self.table.setRowCount(len(self.accountman.accounts['accounts']))

        accounts = self.accountman.get_accounts()
        for i, (k, account) in enumerate(accounts):
            self.table.setRowHeight(i, QtGui.QFontMetrics(
                self.table.font()).height() + 2)
            self.table.setItem(i, 0, AccountItem(k, account['username']))
            self.table.setItem(i, 1, AccountItem(
                k, account['api'], self.icons.get(account['api'])))

        self.table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)

    def select(self, checked):
        if not self.selected_account_num:
            self._error("Please select an account.")
            return
        self.selected.emit(self.selected_account_num,
                           self.remember_chk.isChecked())
        self.close()

    def cancel(self, checked):
        self.aborted.emit()
        self.close()

    def _error(self, msg):
        QMessageBox.critical(self, 'Error', str(msg), QMessageBox.StandardButton.Ok)


class AccountItem(QTableWidgetItem):
    """
    Regular item able to save account item

    """
    num = None

    def __init__(self, num, text, icon=None):
        QTableWidgetItem.__init__(self, text)
        self.num = num
        if icon:
            self.setIcon(icon)


class AccountAddDialog(QDialog):
    def __init__(self, parent, icons, edit=False, username='', password='', api=''):
        QDialog.__init__(self, parent)
        self.edit = edit
        
        self.adding_allow = False
        self.adding_api = None
        self.adding_extra = {}

        # Build UI
        layout = QVBoxLayout()

        formlayout = QFormLayout()
        self.lbl_username = QLabel('Username:')
        self.username = QLineEdit(username)

        pin_layout = QHBoxLayout()
        self.lbl_password = QLabel('Password:')
        self.password = QLineEdit(password)
        self.api = QComboBox()
        self.api.currentIndexChanged.connect(self.s_refresh)
        self.api_auth = QPushButton('Request PIN')
        self.api_auth.clicked.connect(self.s_request_pin)
        pin_layout.addWidget(self.password)
        pin_layout.addWidget(self.api_auth)

        formlayout.addRow(QLabel('Site:'), self.api)
        formlayout.addRow(self.lbl_username, self.username)
        formlayout.addRow(self.lbl_password, pin_layout)

        bottombox = QDialogButtonBox()
        bottombox.addButton(QDialogButtonBox.StandardButton.Save)
        bottombox.addButton(QDialogButtonBox.StandardButton.Cancel)
        bottombox.accepted.connect(self.validate)
        bottombox.rejected.connect(self.reject)

        # Populate APIs
        for libname, lib in sorted(utils.available_libs.items()):
            self.api.addItem(icons[libname], lib[0], libname)

        if self.edit:
            self.username.setEnabled(False)
            self.api.setCurrentIndex(
                self.api.findData(api, QtCore.Qt.ItemDataRole.UserRole))
            self.api.setEnabled(False)

        # Finish layouts
        layout.addLayout(formlayout)
        layout.addWidget(bottombox)

        self.setLayout(layout)

    def validate(self):
        if len(self.username.text()) == 0:
            if len(self.password.text()) == 0:
                self._error('Please fill the credentials fields.')
            else:
                self._error('Please fill the username field.')
        elif len(self.password.text()) == 0:
            self._error('Please fill the password/PIN field.')
        else:
            self.accept()
    
    def s_request_pin(self):
        auth_url = self.adding_api[3]
        if self.adding_api[2] == utils.Login.OAUTH_PKCE:
            self.adding_extra = {'code_verifier': utils.oauth_generate_pkce()}
            auth_url = auth_url % self.adding_extra['code_verifier']
        
        self.adding_allow = True
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(auth_url))
        
    def s_refresh(self, index):
        if not self.edit:
            self.username.setText("")
            self.password.setText("")

        apiname = self.api.itemData(index)
        self.adding_api = utils.available_libs[apiname]
        if self.adding_api[2] in [utils.Login.OAUTH, utils.Login.OAUTH_PKCE]:
            self.lbl_username.setText('Name:')
            self.lbl_password.setText('PIN:')
            self.password.setEchoMode(QLineEdit.EchoMode.Normal)
            self.api_auth.show()
            self.adding_allow = False
        else:
            self.lbl_username.setText('Username:')
            self.lbl_password.setText('Password:')
            self.password.setEchoMode(QLineEdit.EchoMode.Password)
            self.api_auth.hide()
            self.adding_allow = True

    def _error(self, msg):
        QMessageBox.critical(self, 'Error', msg, QMessageBox.StandardButton.Ok)

    @staticmethod
    def do(parent=None, icons=None, edit=False, username='', password='', api='', extra={}):
        dialog = AccountAddDialog(parent, icons, edit, username, password, api)
        result = dialog.exec()

        if result == QDialog.DialogCode.Accepted:
            currentIndex = dialog.api.currentIndex()
            return (
                str(dialog.username.text()),
                str(dialog.password.text()),
                str(dialog.api.itemData(currentIndex)),
                dialog.adding_extra,
            )
        else:
            return None



def get_parser_class(msg, parser_name):
    # Choose the parser we want to use
    if parser_name == 'aie':
        msg.debug('Using AnimeInfoExtractor parser')
        from .animeinfoextractor import AnimeInfoExtractor
        return AnimeInfoExtractor
    elif parser_name == 'anitopy':
        msg.debug('Using Anitopy parser')
        from .anitopy import AnitopyWrapper
        return AnitopyWrapper
    else:
        msg.debug('Unknown parser "{}", falling back to default'.format(parser_name))
        return get_parser_class('aie')



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import os
import random
import re
import shlex
import shutil
import sys
import time
from decimal import Decimal
from functools import lru_cache, partial
from pathlib import Path

from trackma import data
from trackma import messenger
from trackma import utils
from trackma.extras import redirections
from trackma.parser import get_parser_class


class Engine:
    """
    The engine is the controller that handles commands coming from
    the user interface and then queries the Data Handler for the necessary data.
    It doesn't control nor care about how the data is fetched by the Data Handler.

    After instantiating this class, the :func:`start` must be run to initialize the engine.
    Likewise, the :func:`unload` function must be called when you're done using the engine.

    The account and mediatype can be changed later on the fly by calling :func:`reload`.

    The **account** parameter is an account dictionary passed by an Account Manager
    and is used to run the engine.

    The **message_handler** is a reference to a messaging function for the engine
    to send to. Optional.
    """
    data_handler = None
    tracker = None
    redirections = None
    config = {}
    msg = None
    loaded = False
    playing = False
    hooks_available = []

    name = 'Engine'

    signals = {'show_added':        None,
               'show_deleted':      None,
               'episode_changed':   None,
               'score_changed':     None,
               'status_changed':    None,
               'show_synced':       None,
               'sync_complete':     None,
               'queue_changed':     None,
               'playing':           None,
               'prompt_for_update': None,
               'prompt_for_add':    None,
               'tracker_state':     None,
               'episode_missing': None,
               }

    def __init__(self, account=None, message_handler=None, accountnum=None):
        self.msg = messenger.Messenger(message_handler, self.name)

        # Utility parameter to get the account from the account manager
        if accountnum:
            from trackma import accounts
            account = accounts.AccountManager().get_account(accountnum)

        # Initialize
        self._load(account)
        self._init_data_handler()

    def _load(self, account):
        self.account = account

        # Create home directory
        utils.make_dir(utils.to_config_path())
        self.configfile = utils.to_config_path('config.json')

        # Create user directory
        userfolder = "%s.%s" % (account['username'], account['api'])
        utils.make_dir(utils.to_data_path(userfolder))

        self.msg.info('Trackma v{0} - using account {1}({2}).'.format(
            utils.VERSION, account['username'], account['api']))
        self.msg.info('Reading config files...')
        try:
            self.config = utils.parse_config(
                self.configfile, utils.config_defaults)
        except IOError:
            raise utils.EngineFatal("Couldn't open config file.")

        # Expand media directories and ignore those that don't exist
        if isinstance(self.config['searchdir'], str):
            # Compatibility: Turn a string of a single directory into a list
            self.msg.debug("Fixing string searchdir to list.")
            self.config['searchdir'] = [self.config['searchdir']]

        self.searchdirs = [path for path in utils.expand_paths(
            self.config['searchdir']) if self._searchdir_exists(path)]

    def _init_data_handler(self, mediatype=None):
        # Create data handler
        self.data_handler = data.Data(
            self.msg, self.config, self.account, mediatype)
        self.data_handler.connect_signal('show_synced', self._data_show_synced)
        self.data_handler.connect_signal(
            'sync_complete', self._data_sync_complete)
        self.data_handler.connect_signal(
            'queue_changed', self._data_queue_changed)

        # Record the API details
        (self.api_info, self.mediainfo) = self.data_handler.get_api_info()

    def _data_show_synced(self, show, changes):
        self._emit_signal('show_synced', show, changes)

    def _data_sync_complete(self, items):
        self._emit_signal('sync_complete', items)

    def _data_queue_changed(self, queue):
        self._emit_signal('queue_changed', queue)

    def _tracker_detected(self, path, filename):
        self.add_to_library(path, filename)

    def _tracker_removed(self, path, filename):
        self.remove_from_library(path, filename)

    def _tracker_playing(self, showid, playing, episode):
        show = self.get_show_info(showid)
        self._emit_signal('playing', show, playing, episode)

    def _tracker_update(self, show, episode):
        if self.config['tracker_update_prompt']:
            self.msg.info("Prompting for update.")
            self._emit_signal('prompt_for_update', show, episode)
        else:
            try:
                self.set_episode(show['id'], episode)
            except utils.TrackmaError as e:
                self.msg.warn("Can't update episode: {}".format(e))

    def _tracker_unrecognised(self, show, episode):
        if self.config['tracker_not_found_prompt']:
            self._emit_signal('prompt_for_add', show, episode)

    def _tracker_state(self, status):
        self._emit_signal('tracker_state', status)

    def _emit_signal(self, signal, *args):
        try:
            # Call the signal function
            if self.signals[signal]:
                self.signals[signal](*args)
        except AttributeError:
            pass

        # If there are loaded hooks, call the functions in all of them
        for module in self.hooks_available:
            method = getattr(module, signal, None)
            if method is not None:
                self.msg.debug("Calling hook {}:{}...".format(
                    module.__name__, signal))
                try:
                    method(self, *args)
                except Exception as err:
                    self.msg.warn("Exception on hook {}:{}: {}".format(
                        module.__name__, signal, err))

    def _get_tracker_list(self, filter_num=None):
        tracker_list = {}
        if isinstance(filter_num, type(None)):
            source_list = self.get_list()
        elif isinstance(filter_num, list):
            source_list = []
            for status in filter_num:
                if status is not self.mediainfo['statuses_finish']:
                    self.msg.debug("Scanning for "
                                   "{}".format(self.mediainfo['statuses_dict'][status]))
                    source_list = source_list + self.filter_list(status)
        else:
            source_list = self.filter_list(filter_num)

        for show in source_list:
            tracker_list[show['id']] = {
                'id': show['id'],
                'title': show['title'],
                'my_progress': show['my_progress'],
                'total': show['total'],
                'type': None,
                'titles': self.data_handler.get_show_titles(show),
            }

        altnames_map = self.data_handler.get_altnames_map()
        return (tracker_list, altnames_map)

    def _update_tracker(self):
        if self.tracker:
            self.tracker.update_list(self._get_tracker_list())

    def _cleanup(self):
        # If the engine wasn't closed for whatever reason, do it
        if self.loaded:
            self.msg.info("Forcing exit...")
            self.data_handler.unload(True)
            if self.tracker:
                self.tracker.disable()
            self.loaded = False

    def connect_signal(self, signal, callback):
        try:
            self.signals[signal] = callback
        except KeyError:
            raise utils.EngineFatal("Invalid signal.")

    def set_message_handler(self, message_handler):
        """Changes the message handler function on the fly."""
        self.msg = messenger.Messenger(message_handler, self.name)
        self.data_handler.set_message_handler(self.msg)

    def start(self):
        """
        Starts the engine.
        This function should be called before doing anything with the engine,
        as it initializes the data handler.
        """
        if self.loaded:
            raise utils.TrackmaError("Already loaded.")

        self.msg.debug("Starting engine...")

        # Temporary deprecation warning for Python 3.9
        if sys.version_info[:2] == (3, 9):
            self.msg.warn("\n==============="
                          "\nDEPRECATION WARNING: Python 3.9 has reached end of life."
                          "\nTrackma will drop support for it soon. It is recommended"
                          "\nto upgrade to Python 3.10 or newer."
                          "\n===============")

        # Start the data handler
        try:
            (self.api_info, self.mediainfo) = self.data_handler.start()
        except utils.DataError as e:
            raise utils.DataFatal(str(e))
        except utils.APIError as e:
            raise utils.APIFatal(str(e))

        # Load redirection file if supported
        api = self.api_info['shortname']
        mediatype = self.data_handler.userconfig['mediatype']
        if redirections.supports(api, mediatype):
            if utils.file_exists(utils.to_config_path('anime-relations.txt')):
                fname = utils.to_config_path('anime-relations.txt')
                self.msg.debug("Using user-provided redirection file.")
            else:
                fname = utils.to_data_path('anime-relations.txt')
                if self.config['redirections_time'] and (
                        not utils.file_exists(fname) or
                        utils.file_older_than(fname, self.config['redirections_time'] * 86400)):
                    self.msg.info("Syncing redirection file...")
                    self.msg.debug("Syncing from: %s" %
                                   self.config['redirections_url'])
                    utils.sync_file(fname, self.config['redirections_url'])

            if not utils.file_exists(fname):
                self.msg.debug("Defaulting to repo provided redirections file.")
                fname = utils.DATADIR + '/anime-relations/anime-relations.txt'

            self.msg.info("Parsing redirection file...")
            try:
                self.redirections = redirections.parse_anime_relations(
                    fname, api)
            except Exception as e:
                self.msg.warn("Error parsing anime-relations.txt!")
                self.msg.debug("{}".format(e))

        # Determine parser library
        try:
            self.msg.debug(self.name, "Initializing parser...")
            self.parser_class = get_parser_class(self.msg, self.config['title_parser'])
        except ImportError as e:
            self.msg.warn(self.name, "Couldn't import specified parser: {}; {}".format(
                self.config['title_parser'], e))
            self.msg.warn(self.name, "Falling back to aie...")
            self.parser_class = get_parser_class(self.msg, "aie")

        # Rescan library if necessary
        if self.config['library_autoscan']:
            try:
                self.scan_library()
            except utils.TrackmaError as e:
                self.msg.warn("Can't auto-scan library: {}".format(e))

        # Load hook files
        if self.config['use_hooks']:
            hooks_dir = utils.to_config_path('hooks')
            if os.path.isdir(hooks_dir):
                import importlib.util
                import pkgutil

                self.msg.info("Importing user hooks...")
                for finder, name, _ in pkgutil.iter_modules([hooks_dir]):
                    # List all the hook files in the hooks folder, import them
                    # and call the init() function if they have them
                    # We build the list "hooks available" with the loaded modules
                    # for later calls.
                    try:
                        self.msg.debug("Importing hook {}...".format(name))
                        module_spec = finder.find_spec(name)
                        module = importlib.util.module_from_spec(module_spec)
                        module_spec.loader.exec_module(module)
                        if hasattr(module, 'init'):
                            module.init(self)
                        self.hooks_available.append(module)
                    except ImportError:
                        self.msg.warn("Error importing hook {}.".format(name))
                        self.msg.exception(sys.exc_info())

        # Start tracker
        if self.mediainfo.get('can_play') and self.config['tracker_enabled']:
            self.msg.debug("Initializing tracker...")
            try:
                TrackerClass = self._get_tracker_class(
                    self.config['tracker_type'])

                self.tracker = TrackerClass(self.msg,
                                            self._get_tracker_list(),
                                            self.config,
                                            self.searchdirs,
                                            self.redirections,
                                            )
                self.tracker.connect_signal('detected', self._tracker_detected)
                self.tracker.connect_signal('removed', self._tracker_removed)
                self.tracker.connect_signal('playing', self._tracker_playing)
                self.tracker.connect_signal('update', self._tracker_update)
                self.tracker.connect_signal(
                    'unrecognised', self._tracker_unrecognised)
                self.tracker.connect_signal('state', self._tracker_state)
            except ImportError:
                self.msg.warn("Couldn't import specified tracker: {}".format(
                    self.config['tracker_type']))
                self.msg.exception(sys.exc_info())

        self.loaded = True
        self.msg.debug("Engine started")
        return True

    def unload(self):
        """
        Closes the data handler and closes the engine cleanly.
        This should be called when closing the client application, or when you're
        sure you're not going to use the engine anymore. This does all the necessary
        procedures to close the data handler cleanly and then itself.

        """
        if self.loaded:
            self.msg.info("Unloading...")
            self.data_handler.unload()
            if self.tracker:
                self.tracker.disable()

            # If there are loaded hooks, unload them
            self.msg.info("Unloading user hooks...")
            for module in self.hooks_available.copy():
                self.msg.debug("Unloading hook {}...".format(
                    module.__name__))
                try:
                    if hasattr(module, 'destroy'):
                        module.destroy(self)
                    self.hooks_available.remove(module)
                except Exception as err:
                    self.msg.warn("Error destroying hook {}: {}".format(
                        module.__name__, err))

            self.loaded = False

    def reload(self, account=None, mediatype=None):
        """Changes the API and/or mediatype and reloads itself."""
        if self.loaded:
            self.unload()

        if account:
            self._load(account)

        self._init_data_handler(mediatype)
        self.start()

    def get_config(self, key):
        """Returns the specified key from the configuration."""
        return self.config[key]

    def get_userconfig(self, key):
        return self.data_handler.userconfig[key]

    def set_config(self, key, value):
        """
        Writes the defined key to the configuration.
        Note that this writes the configuration only to memory; when you're
        done doing all necessary changes, make sure to write the configuration file
        with :func:`save_config`."""
        self.config[key] = value

    def save_config(self):
        """Writes all configuration files to disk."""

        # Save config file
        utils.save_config(self.config, self.configfile)

    def get_list(self):
        """
        Returns the full show list requested from the data handler as a list of show dictionaries.
        If you only need shows in a specified status, use :func:`filter_list`.
        """
        return self.data_handler.get().values()

    def get_show_info(self, showid=None, title=None, filename=None):
        """
        Returns the show dictionary for the specified **showid**.
        """
        showdict = self.data_handler.get()

        if showid:
            # Get show by ID
            try:
                return showdict[showid]
            except KeyError:
                raise utils.EngineError("Show not found.")
        elif title:
            showdict = self.data_handler.get()
            # Get show by title, slower
            for show in showdict.values():
                if show['title'] == title:
                    return show
            raise utils.EngineError("Show not found.")
        elif filename:
            # Guess show by filename
            self.msg.debug("Guessing by filename.")

            anime_info = self.parser_class(self.msg, filename)
            (show_title, ep) = anime_info.getName(), anime_info.getEpisode()
            self.msg.debug("Show guess: {}".format(show_title))

            if show_title:
                tracker_list = self._get_tracker_list()

                show = utils.guess_show(show_title, tracker_list)
                if show:
                    return utils.redirect_show((show, ep), self.redirections, tracker_list)
                else:
                    raise utils.EngineError("Show not found.")
            else:
                raise utils.EngineError("File name not recognized.")

    def get_show_details(self, show):
        """
        Returns detailed information about **show** requested from the data handler.
        """
        return self.data_handler.info_get(show)

    def regex_list(self, regex):
        """
        It asks the data handler to do a regex search for a show and returns the
        list of show dictionaries with all the matches.
        """
        showlist = self.data_handler.get()
        return list(v for k, v in showlist.items() if re.search(regex, v['title'], re.I))

    def regex_list_titles(self, pattern):
        # TODO : Temporal hack for the client autocomplete function
        showlist = self.data_handler.get()
        newlist = list()
        for v in showlist.values():
            if re.match(pattern, v['title'], re.I):
                if ' ' in v['title']:
                    newlist.append('"' + v['title'] + '" ')
                else:
                    newlist.append(v['title'] + ' ')

        return newlist

    def tracker_status(self):
        """
        Asks the tracker for its current status.
        """

        if self.tracker:
            return self.tracker.get_status()

        return None

    def search(self, criteria, method=utils.SearchMethod.KW):
        """
        Request a remote list of shows matching the criteria
        and returns it as a list of show dictionaries.
        This is useful to add a show.
        """
        if method not in self.mediainfo.get('search_methods', [utils.SearchMethod.KW]):
            raise utils.EngineError(
                'Search method not supported by API or mediatype.')

        return self.data_handler.search(criteria, method)

    def add_show(self, show, status=None):
        """
        Adds **show** to the list and queues the list update
        for the next sync.
        """
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_add'):
            raise utils.EngineError('Operation not supported by API.')

        # Set to the requested status
        if status:
            if status not in self.mediainfo['statuses']:
                raise utils.EngineError('Invalid status.')

            show['my_status'] = status

        # Add in data handler
        self.data_handler.queue_add(show)

        # Update the tracker with the new information
        self._update_tracker()

        # Emit signal
        self._emit_signal('show_added', show)

    def set_episode(self, showid, newep):
        """
        Updates the progress of the specified **showid** to **newep**
        and queues the list update for the next sync.
        """
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_update'):
            raise utils.EngineError('Operation not supported by API.')

        # Check for the episode number
        try:
            newep = int(newep)
        except ValueError:
            raise utils.EngineError('Episode must be numeric.')

        # Get the show info
        show = self.get_show_info(showid)
        # More checks
        if (show['total'] and newep > show['total']) or newep < 0:
            raise utils.EngineError('Episode out of limits.')
        if show['my_progress'] == newep:
            raise utils.EngineError("Show already at episode %d" % newep)

        # Change episode
        self.msg.info("Updating show %s to episode %d..." %
                      (show['title'], newep))
        self.data_handler.queue_update(show, 'my_progress', newep)

        # Emit signal
        self._emit_signal('episode_changed', show)

        # Change status if required
        oldstatus = show['my_status']
        if self.config['auto_status_change'] and self.mediainfo.get('can_status'):
            try:
                if newep == show['total'] and self.mediainfo.get('statuses_finish'):
                    if (
                            not self.config['auto_status_change_if_scored'] or
                            not self.mediainfo.get('can_score') or
                            show['my_score']
                    ):
                        # Change to finished status
                        self.set_status(
                            show['id'], self._guess_new_finish(show))
                    else:
                        self.msg.warn("Updated episode but status won't be changed until a score is set.")
                elif newep == 1 and self.mediainfo.get('statuses_start'):
                    # Change to start status
                    self.set_status(show['id'], self._guess_new_start(show))
            except utils.EngineError as e:
                # Only warn about engine errors since status change here is not critical
                self.msg.warn('Updated episode but status wasn\'t changed: %s' % e)

        # Change dates if required
        if self.config['auto_date_change'] and self.mediainfo.get('can_date'):
            start_date = finish_date = None

            try:
                initial_status = self.mediainfo.get('statuses_start')[0]

                if newep == 1 and show['my_status'] == initial_status:
                    start_date = datetime.date.today()
                if newep == show['total'] and oldstatus == initial_status:
                    finish_date = datetime.date.today()

                self.set_dates(show['id'], start_date, finish_date)
            except utils.EngineError as e:
                # Only warn about engine errors since date change here is not critical
                self.msg.warn('Updated episode but dates weren\'t changed: %s' % e)

        # Update the tracker with the new information
        self._update_tracker()

        return show

    def set_dates(self, showid, start_date=None, finish_date=None):
        """
        Updates the start date and finish date of a show.
        If any of the two are None, it won't be changed.
        """
        if not self.mediainfo.get('can_date'):
            raise utils.EngineError('Operation not supported by API.')

        show = self.get_show_info(showid)

        # Change the start date if required
        if start_date:
            if not isinstance(start_date, datetime.date):
                raise utils.EngineError('start_date must be a Date object.')
            self.msg.info("Updating show %s start date to %s..." %
                      (show['title'], start_date))
            self.data_handler.queue_update(show, 'my_start_date', start_date)

        if finish_date:
            if not isinstance(finish_date, datetime.date):
                raise utils.EngineError('finish_date must be a Date object.')
            self.msg.info("Updating show %s finish date to %s..." %
                      (show['title'], finish_date))
            self.data_handler.queue_update(show, 'my_finish_date', finish_date)

    def set_score(self, showid, newscore):
        """
        Updates the score of the specified **showid** to **newscore**
        and queues the list update for the next sync.
        """
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_score'):
            raise utils.EngineError('Operation not supported by API.')

        # Check for the correctness of the score
        if (Decimal(str(newscore)) % Decimal(str(self.mediainfo['score_step']))) != 0:
            raise utils.EngineError('Invalid score.')

        # Convert to proper type
        if isinstance(self.mediainfo['score_step'], int):
            newscore = int(newscore)
        else:
            newscore = float(newscore)

        # Get the show and update it
        show = self.get_show_info(showid)
        # More checks
        if newscore > self.mediainfo['score_max']:
            raise utils.EngineError('Score out of limits.')
        if show['my_score'] == newscore:
            raise utils.EngineError("Score already at %s" % newscore)

        # Change score
        self.msg.info("Updating show %s score to %s..." %
                      (show['title'], newscore))
        self.data_handler.queue_update(show, 'my_score', newscore)

        # Emit signal
        self._emit_signal('score_changed', show)

        # Change status if required
        if (
                show['total'] and
                show['my_progress'] == show['total'] and
                show['my_score'] and
                self.mediainfo.get('can_status') and
                self.config['auto_status_change'] and
                self.config['auto_status_change_if_scored'] and
                self.mediainfo.get('statuses_finish')
        ):
            try:
                self.set_status(show['id'], self._guess_new_finish(show))
            except utils.EngineError as e:
                # Only warn about engine errors since status change here is not critical
                self.msg.warn('Updated episode but status wasn\'t changed: %s' % e)

        return show

    def set_status(self, showid, newstatus):
        """
        Updates the score of the specified **showid** to **newstatus** (number)
        and queues the list update for the next sync.
        """
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_status'):
            raise utils.EngineError('Operation not supported by API.')

        try:
            newstatus = int(newstatus)
        except ValueError:
            pass  # It's not necessary for it to be an int

        # Check if the status is valid
        _statuses = self.mediainfo['statuses_dict']
        if newstatus not in _statuses:
            raise utils.EngineError('Invalid status.')

        # Get the show and update it
        show = self.get_show_info(showid)
        # More checks
        if show['my_status'] == newstatus:
            raise utils.EngineError("Show already in %s." %
                                    _statuses[newstatus])

        # Change status
        old_status = show['my_status']
        self.msg.info("Updating show %s status to %s..." %
                      (show['title'], _statuses[newstatus]))
        self.data_handler.queue_update(show, 'my_status', newstatus)

        # Emit signal
        self._emit_signal('status_changed', show, old_status)

        return show

    def set_tags(self, showid, newtags):
        """
        Updates the tags of the specified **showid** to **newtags**
        and queues the list update for the next sync.
        """
        # Check if operation is supported by the API
        if 'can_tag' not in self.mediainfo or not self.mediainfo.get('can_tag'):
            raise utils.EngineError('Operation not supported by API.')

        # Get the show and update it
        show = self.get_show_info(showid)
        # More checks
        if show['my_tags'] == newtags:
            raise utils.EngineError("Tags already %s" % newtags)

        # Change score
        self.msg.info("Updating show %s to tags '%s'..." %
                      (show['title'], newtags))
        self.data_handler.queue_update(show, 'my_tags', newtags)

        # Emit signal
        self._emit_signal('tags_changed', show)

        return show

    def delete_show(self, show):
        """
        Deletes **show** completely from the list and queues the list update for the next sync.
        """
        if not self.mediainfo.get('can_delete'):
            raise utils.EngineError('Operation not supported by API.')

        # Add in data handler
        self.data_handler.queue_delete(show)

        # Update the tracker with the new information
        self._update_tracker()

        # Emit signal
        self._emit_signal('show_deleted', show)

    def library(self):
        return self.data_handler.library_get()

    def scan_library(self, my_status=None, rescan=False, path=None):
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_play'):
            raise utils.EngineError(
                'Operation not supported by current site or mediatype.')
        if not self.config['searchdir']:
            raise utils.EngineError('Media directories not set.')

        t = time.time()
        library = {}
        library_cache = self.data_handler.library_cache_get()

        if not my_status:
            if self.config['scan_whole_list']:
                my_status = self.mediainfo['statuses']
            else:
                my_status = self.mediainfo.get(
                    'statuses_library', self.mediainfo['statuses_start'])

        if rescan:
            self.msg.info("Scanning local library (overriding cache)...")
        else:
            self.msg.info("Scanning local library...")

        tracker_list = self._get_tracker_list(my_status)
        guess_show = lru_cache(partial(utils.guess_show, tracker_list=tracker_list))

        paths = [path] if path else self.searchdirs
        for searchdir in paths:
            self.msg.debug("Directory: %s" % searchdir)

            # Do a full listing of the media directory
            for fullpath, filename in utils.regex_find_videos(searchdir):
                if self.config['library_full_path']:
                    filename = self._get_relative_path_or_basename(searchdir, fullpath)
                (library, library_cache) = self._add_show_to_library(
                    library, library_cache, rescan, fullpath, filename, tracker_list, guess_show)

            self.msg.debug("Time: %s" % (time.time() - t))
            self.data_handler.library_save(library)
            self.data_handler.library_cache_save(library_cache)
        return library

    def remove_from_library(self, path, filename):
        library = self.data_handler.library_get()
        library_cache = self.data_handler.library_cache_get()
        tracker_list = self._get_tracker_list()
        fullpath = path+"/"+filename
        # Only remove if the filename matches library entry
        if filename in library_cache and library_cache[filename]:
            (show_id, show_ep) = library_cache[filename]
            if show_id and show_ep and show_id \
                    and library.get(show_id, {}).get(show_ep) == fullpath:
                self.msg.debug("File removed from local library: %s" % fullpath)
                library_cache.pop(filename, None)
                library[show_id].pop(show_ep, None)

    def add_to_library(self, path, filename, rescan=False):
        # The inotify tracker tells us when files are created in
        # or moved within our library directory, so we call this.
        library = self.data_handler.library_get()
        library_cache = self.data_handler.library_cache_get()
        tracker_list = self._get_tracker_list()
        fullpath = path+"/"+filename
        guess_show = partial(utils.guess_show, tracker_list=tracker_list)
        self._add_show_to_library(
            library, library_cache, rescan, fullpath, filename, tracker_list, guess_show)

    def _add_show_to_library(self, library, library_cache, rescan, fullpath, filename, tracker_list, guess_show):
        show_id = None
        if not rescan and filename in library_cache:
            # If the filename was already seen before
            # use the cached information, if there's no information (None)
            # then it means it doesn't correspond to any show in the list
            # and can be safely skipped.
            if library_cache[filename]:
                (show_id, show_ep) = library_cache[filename]
                if type(show_ep) is tuple:
                    (show_ep_start, show_ep_end) = show_ep
                else:
                    show_ep_start = show_ep_end = show_ep
            else:
                return library, library_cache
        else:
            # If the filename has not been seen, extract
            # the information from the filename and do a fuzzy search
            # on the user's list. Cache the information.
            # If it fails, cache it as None.
            anime_info = self.parser_class(self.msg, filename)
            show_title = anime_info.getName()
            (show_ep_start, show_ep_end) = anime_info.getEpisodeNumbers(True)
            if show_title:
                show = guess_show(show_title)
                if show:
                    self.msg.debug("Adding to library: {}".format(fullpath))
                    self.msg.debug("Show guess: {}".format(show_title))

                    if show_ep_start == show_ep_end:
                        # TODO : Support redirections for episode ranges
                        (show, show_ep) = utils.redirect_show(
                            (show, show_ep_start), self.redirections, tracker_list)
                        show_ep_end = show_ep_start = show_ep

                        self.msg.debug("Redirected to: {} - {}".format(
                            show['title'], show_ep))
                        library_cache[filename] = (show['id'], show_ep)
                    else:
                        library_cache[filename] = (
                            show['id'], (show_ep_start, show_ep_end))

                    show_id = show['id']
                else:
                    self.msg.debug("Unable to match '{}', skipping: {}"
                                   .format(show_title, fullpath))
                    library_cache[filename] = None
            else:
                self.msg.debug("Not recognized, skipping: {}".format(fullpath))
                library_cache[filename] = None

        # After we got our information, add it to our library
        if show_id:
            if show_id not in library:
                library[show_id] = {}
            for show_ep in range(show_ep_start, show_ep_end+1):
                library[show_id][show_ep] = fullpath

        return library, library_cache

    def get_episode_path(self, show, episode=0):
        """
        This function returns the full path of the requested episode from the requested show.
        If the episode is unspecified, it will return any episode from the requested show.
        """
        library = self.library()
        showid = show['id']

        if showid not in library:
            raise utils.EngineError('Show not in library.')

        # Get the specified episode, otherwise get any if unspecified
        if episode:
            if episode not in library[showid]:
                raise utils.EngineError('Episode not in library.')
            return library[showid][episode]
        else:
            return next(iter(library[showid].values()))

    def play_random(self):
        """
        This function will pick a random show that has a new episode to watch
        and return the arguments to play it.
        """
        library = self.library()
        newep = []

        self.msg.info('Looking for random episode.')

        for showid, eps in library.items():
            try:
                show = self.get_show_info(showid)
            except utils.EngineError:
                continue # In library but not available
            if show['my_progress'] + 1 in eps:
                newep.append(show)

        if not newep:
            raise utils.EngineError('No new episodes found to pick from.')

        show = random.choice(newep)
        return self.play_episode(show)

    def play_episode(self, show, playep=0):
        """
        Does a local search in the hard disk (in the folder specified by the config file)
        for the specified episode (**playep**) for the specified **show**.

        If no **playep** is specified, the next episode of the show will be returned.
        """
        # Check if operation is supported by the API
        if not self.mediainfo.get('can_play'):
            raise utils.EngineError(
                'Operation not supported by current site or mediatype.')

        try:
            playep = int(playep)
        except ValueError:
            raise utils.EngineError('Episode must be numeric.')

        if not show:
            raise utils.EngineError('Show given is invalid')

        if playep <= 0:
            playep = show['my_progress'] + 1

        if show['total'] and playep > show['total']:
            raise utils.EngineError('Episode beyond limits.')

        self.msg.info("Getting '%s' episode '%s' from library..." %
                        (show['title'], playep))

        try:
            filename = self.get_episode_path(show, playep)
        except utils.EngineError:
            self.msg.info("Episode not found. Calling hooks...")
            self._emit_signal("episode_missing", show, playep)
            return []

        self.msg.info('Found. Starting player...')
        args = shlex.split(self.config['player'])

        if not args:
            raise utils.EngineError('Player not set up, check your config.json')

        args[0] = shutil.which(args[0])

        if not args[0]:
            raise utils.EngineError('Player not found, check your config.json')

        args.append(filename)
        return args

    def open_show_folder(self, show_id):
        show = self.get_show_info(show_id)
        filename = self.get_episode_path(show)
        try:
            utils.open_folder(os.path.dirname(filename))
        except OSError:
            raise utils.EngineError("Could not open folder.")

    def queue_clear(self):
        """Clears the data handler queue and discards any unsynced change."""
        return self.data_handler.queue_clear()

    def altname(self, showid, newname=None):
        """
        If **newname** is specified, it gets the alternate name of **showid**.
        Otherwise, it sets the alternate name of **showid** to **newname**.
        """
        if newname is not None:
            if newname == '':
                self.data_handler.altname_clear(showid)
                self.msg.info('Cleared alternate name.')
            else:
                self.data_handler.altname_set(showid, newname)
                self.msg.info('Changed alternate name to %s.' % newname)
            # Update the tracker with the new altname
            self._update_tracker()
        else:
            return self.data_handler.altname_get(showid)

    def altnames(self):
        """
        Gets a dictionary of all set alternative names.
        """
        return self.data_handler.altnames_get()

    def filter_list(self, status_num):
        """
        Returns a show list with the shows in the specified **status_num** status.
        If you need a list with all the shows, use :func:`get_list`.
        """
        showlist = self.data_handler.get()
        return list(v for k, v in showlist.items() if v['my_status'] == status_num)

    def list_download(self):
        """Asks the data handler to download the remote list."""
        self.data_handler.queue_clear()
        self.data_handler.download_data()
        self._update_tracker()

    def list_upload(self):
        """Asks the data handler to upload the unsynced changes in the queue."""
        self.data_handler.process_queue()
        # for show in result:
        #    self._emit_signal('episode_changed', show)

    def get_queue(self):
        """Asks the data handler for the items in the current queue."""
        return self.data_handler.queue

    def _get_relative_path_or_basename(self, searchdir, fullpath):
        """Determine the path relative to a directory or the basename if not a sub-path.

        Used for including the folder name(s) for show detection.
        """
        path = Path(fullpath)
        try:
            return str(path.relative_to(searchdir))
        except ValueError:
            return path.basename

    def _searchdir_exists(self, path):
        """Variation of dir_exists that warns the user if the path doesn't exist."""
        if not utils.dir_exists(path):
            self.msg.warn("The specified media directory {} doesn't exist!".format(path))
            return False
        return True

    def _guess_new_finish(self, show):
        try:
            # Use corresponding finish status if we're already in a start status
            new_index = self.mediainfo['statuses_start'].index(
                show['my_status'])
            new_status = self.mediainfo['statuses_finish'][new_index]
        except ValueError:
            new_status = self.mediainfo['statuses_finish'][0]
        except IndexError:
            new_status = self.mediainfo['statuses_finish'][-1]

        return new_status

    def _guess_new_start(self, show):
        try:
            # Use following start status if we're already in a finish status
            new_index = self.mediainfo['statuses_finish'].index(
                show['my_status'])
            new_status = self.mediainfo['statuses_start'][new_index+1]
        except ValueError:
            new_status = self.mediainfo['statuses_start'][0]
        except IndexError:
            new_status = self.mediainfo['statuses_start'][-1]

        return new_status

    def _get_tracker_class(self, ttype):
        # Choose the tracker we want to start
        if ttype == 'plex':
            from trackma.tracker.plex import PlexTracker
            return PlexTracker
        if ttype == 'jellyfin':
            from trackma.tracker.jellyfin import JellyfinTracker
            return JellyfinTracker
        elif ttype == 'kodi':
            from trackma.tracker.kodi import KodiTracker
            return KodiTracker
        elif ttype == 'mpris':
            from trackma.tracker.mpris import MprisTracker
            return MprisTracker
        elif ttype == 'inotify_auto':
            try:
                return self._get_tracker_class('pyinotify')
            except ImportError:
                return self._get_tracker_class('inotify')
        elif ttype == 'pyinotify':
            from trackma.tracker.pyinotify import pyinotifyTracker
            return pyinotifyTracker
        elif ttype == 'inotify':
            from trackma.tracker.inotify import inotifyTracker
            return inotifyTracker
        elif ttype == 'win32':
            from trackma.tracker.win32 import Win32Tracker
            return Win32Tracker
        elif ttype == 'polling':
            from trackma.tracker.polling import PollingTracker
            return PollingTracker
        else:
            # Guess the working tracker
            if os.name == 'nt':
                return self._get_tracker_class('win32')

            # Try trackers in this order: pyinotify, inotify, polling
            try:
                return self._get_tracker_class('inotify_auto')
            except ImportError:
                return self._get_tracker_class('polling')
